"""

neuron
======

For empirically-based simulations of neurons and networks of neurons in Python.

This is the top-level module of the official python interface to
the NEURON simulation environment (http://neuron.yale.edu/neuron/).

Documentation is available in the docstrings.

For a list of available names, try dir(neuron).

Example:

$ ipython
In [1]: import neuron
NEURON -- VERSION 6.2 2008-08-22
Duke, Yale, and the BlueBrain Project -- Copyright 1984-2007
See http://neuron.yale.edu/credits.html


In [2]: neuron.h ?






Important names and sub-packages
---------------------

For help on these useful functions, see their docstrings:

  load_mechanisms


neuron.h

   The top-level Hoc interpreter.

   Execute Hoc commands by calling h with a string argument:

   >>> h('objref myobj')
   >>> h('myobj = new Vector(10)')

   All Hoc defined variables are accessible by attribute access to h.

   Example:

   >>> print h.myobj.x[9]

   Hoc Classes are also defined, for example:

   >>> v = h.Vector([1,2,3])
   >>> soma = h.Section()

   More help is available for the respective class by looking in the object docstring:

   >>> help(h.Vector)



neuron.gui

   Import this package if you are using NEURON as an extension to Python,
   and you would like to use the NEURON GUI.

   If you are using NEURON with embedded python, "nrniv -python",
   use rather "nrngui -python" if you would like to use the NEURON GUI.

$Id: __init__.py,v 1.1 2008/05/26 11:39:44 hines Exp hines $

"""

## With Python launched under Linux, shared libraries are apparently imported
## using RTLD_LOCAL. For --with-paranrn=dynamic, this caused a failure when
## libnrnmpi.so is dynamically loaded because nrnmpi_myid (and other global
## variables in src/nrnmpi/nrnmpi_def_cinc) were not resolved --- even though
## all those variables are defined in src/oc/nrnmpi_dynam.c and that
## does a dlopen("libnrnmpi.so", RTLD_NOW | RTLD_GLOBAL) .
## In this case setting the dlopenflags below fixes the problem. But it
## seems that DLFCN is often not available.
## This situation is conceptually puzzling because there
## never seems to be a problem dynamically loading libnrnmech.so, though it
## obviously makes use of many names in the rest of NEURON. Anyway,
## we make the following available in case it is ever needed at least to
## verify that some import problem is traceable to this issue.
## The problem can be resolved in two ways. 1) see src/nrnmpi/nrnmpi_dynam.c
## which promotes liboc.so and libnrniv.so to RTLD_GLOBAL  (commented out).
## 2) The better way of specifying those libraries to libnrnmpi_la_LIBADD
## in src/nrnmpi/Makefile.am . This latter also explains why libnrnmech.so
## does not have this problem.
# try:
#  import sys
#  import DLFCN
#  sys.setdlopenflags(DLFCN.RTLD_NOW | DLFCN.RTLD_GLOBAL)
# except:
#  pass

import sys
import os
import warnings
import weakref

embedded = True if "hoc" in sys.modules else False

try:  # needed since python 3.8 on windows if python launched
    # do this here as NEURONHOME may be changed below
    nrnbindir = os.path.abspath(os.environ["NEURONHOME"] + "/bin")
    os.add_dll_directory(nrnbindir)
except:
    pass

# With pip we need to rewrite the NEURONHOME
nrn_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".data/share/nrn"))
if os.path.isdir(nrn_path):
    os.environ["NEURONHOME"] = nrn_path

# On OSX, dlopen might fail if not using full library path
try:
    from sys import platform

    if platform == "darwin":
        from ctypes.util import find_library

        mpi_library_path = find_library("mpi")
        if mpi_library_path and "MPI_LIB_NRN_PATH" not in os.environ:
            os.environ["MPI_LIB_NRN_PATH"] = mpi_library_path
except:
    pass

try:
    from . import hoc
except:
    import neuron.hoc

import nrn
import _neuron_section

h = hoc.HocObject()
version = h.nrnversion(5)
__version__ = version
_original_hoc_file = None
if not hasattr(hoc, "__file__"):
    # first try is to derive from neuron.__file__
    origin = None  # path to neuron/__init__.py
    from importlib import util

    mspec = util.find_spec("neuron")
    if mspec:
        origin = mspec.origin
    if origin is not None:
        import sysconfig

        hoc_path = (
            origin.rstrip("__init__.py")
            + "hoc"
            + sysconfig.get_config_var("EXT_SUFFIX")
        )
        setattr(hoc, "__file__", hoc_path)
else:
    _original_hoc_file = hoc.__file__
# As a workaround to importing doc at neuron import time
# (which leads to chicken and egg issues on some platforms)
# define a dummy help function which imports doc,
# calls the real help function, and reassigns neuron.help to doc.help
# (thus replacing the dummy)
def help(request=None):
    global help
    from neuron import doc

    doc.help(request)
    help = doc.help


try:
    import pydoc

    pydoc.help = help
except:
    pass

# Global test-suite function


def test(exitOnError=True):
    """Runs a global battery of unit tests on the neuron module."""
    import neuron.tests
    import unittest

    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(neuron.tests.suite()).wasSuccessful()
    if exitOnError and result is False:
        sys.exit(1)
    return result


def test_rxd(exitOnError=True):
    """Runs a tests on the rxd and crxd modules."""
    import neuron.tests
    import unittest

    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(neuron.tests.test_rxd.suite()).wasSuccessful()
    if exitOnError and result is False:
        sys.exit(1)
    return result


# ------------------------------------------------------------------------------
# class factory for subclassing h.anyclass
# h.anyclass methods may be overridden. If so the base method can be called
# using the idiom self.basemethod = self.baseattr('methodname')
# ------------------------------------------------------------------------------

import sys, types
from neuron.hclass3 import HocBaseObject, hclass

# global list of paths already loaded by load_mechanisms
nrn_dll_loaded = []


def load_mechanisms(path, warn_if_already_loaded=True):
    """
    load_mechanisms(path)

    Search for and load NMODL mechanisms from the path given.

    This function will not load a mechanism path twice.

    The path should specify the directory in which nrnivmodl or mknrndll was run,
    and in which the directory 'i686' (or 'x86_64' or 'powerpc' depending on your platform)
    was created"""

    import platform

    global nrn_dll_loaded
    if path in nrn_dll_loaded:
        if warn_if_already_loaded:
            print("Mechanisms already loaded from path: %s.  Aborting." % path)
        return True

    # in case NEURON is assuming a different architecture to Python,
    # we try multiple possibilities

    libname = "libnrnmech.so"
    libsubdir = ".libs"
    arch_list = [platform.machine(), "i686", "x86_64", "powerpc", "umac"]

    # windows loads nrnmech.dll
    if h.unix_mac_pc() == 3:
        libname = "nrnmech.dll"
        libsubdir = ""
        arch_list = [""]

    for arch in arch_list:
        lib_path = os.path.join(path, arch, libsubdir, libname)
        if os.path.exists(lib_path):
            h.nrn_load_dll(lib_path)
            nrn_dll_loaded.append(path)
            return True
    print("NEURON mechanisms not found in %s." % path)
    return False


import os, sys

if "NRN_NMODL_PATH" in os.environ:
    nrn_nmodl_path = os.environ["NRN_NMODL_PATH"].split(":")
    print("Auto-loading mechanisms:")
    print("NRN_NMODL_PATH=%s" % os.environ["NRN_NMODL_PATH"])
    for x in nrn_nmodl_path:
        # print "from path %s:" % x
        load_mechanisms(x)
        # print "\n"
    print("Done.\n")


# ------------------------------------------------------------------------------
# Python classes and functions without a Hoc equivalent, mainly for internal
# use within this file.
# ------------------------------------------------------------------------------


class HocError(Exception):
    pass


class Wrapper(object):
    """Base class to provide attribute access for HocObjects."""

    def __getattr__(self, name):
        if name == "hoc_obj":
            return self.__dict__["hoc_obj"]
        else:
            try:
                return self.__getattribute__(name)
            except AttributeError:
                return self.hoc_obj.__getattribute__(name)

    def __setattr__(self, name, value):
        try:
            self.hoc_obj.__setattr__(name, value)
        except LookupError:
            object.__setattr__(self, name, value)


def new_point_process(name, doc=None):
    """
    Returns a Python-wrapped hoc class where the object needs to be associated
    with a section.

    doc - specify a docstring for the new pointprocess class
    """
    h("obfunc new_%s() { return new %s($1) }" % (name, name))

    class someclass(Wrapper):
        __doc__ = doc

        def __init__(self, section, position=0.5):
            assert 0 <= position <= 1
            section.push()
            self.__dict__["hoc_obj"] = getattr(h, "new_%s" % name)(
                position
            )  # have to put directly in __dict__ to avoid infinite recursion with __getattr__
            h.pop_section()

    someclass.__name__ = name
    return someclass


def new_hoc_class(name, doc=None):
    """
    Returns a Python-wrapped hoc class where the object does not need to be
    associated with a section.

    doc - specify a docstring for the new hoc class
    """
    h("obfunc new_%s() { return new %s() }" % (name, name))

    class someclass(Wrapper):
        __doc__ = doc

        def __init__(self, **kwargs):
            self.__dict__["hoc_obj"] = getattr(h, "new_%s" % name)()
            for k, v in list(kwargs.items()):
                setattr(self.hoc_obj, k, v)

    someclass.__name__ = name
    return someclass


# ------------------------------------------------------------------------------
# Python equivalents to Hoc functions
# ------------------------------------------------------------------------------


def xopen(*args, **kwargs):
    """
    Syntax:
        ``neuron.xopen("hocfile")``


        ``neuron.xopen("hocfile", "RCSrevision")``


    Description:
        ``h.xopen()`` executes the commands in ``hocfile``.  This is a convenient way
        to define user functions and procedures.
        An optional second argument is the RCS revision number in the form of a
        string. The RCS file with that revision number is checked out into a
        temporary file and executed. The temporary file is then removed.  A file
        of the same primary name is unaffected.

    This function is deprecated and will be removed in a future release.
    Use ``h.xopen`` instead.
    """
    warnings.warn(
        "neuron.xopen is deprecated; use h.xopen instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return h.xopen(*args, **kwargs)


def quit(*args, **kwargs):
    """
    Exits the program. Can be used as the action of a button. If edit buffers
    are open you will be asked if you wish to save them before the final exit.

    This function is deprecated and will be removed in a future release.
    Use ``h.quit()`` or ``sys.exit()`` instead. (Note: sys.exit will not prompt
    for saving edit buffers.)
    """
    warnings.warn(
        "neuron.quit() is deprecated; use h.quit() or sys.exit() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return h.quit(*args, **kwargs)


def hoc_execute(hoc_commands, comment=None):
    assert isinstance(hoc_commands, list)
    if comment:
        logging.debug(comment)
    for cmd in hoc_commands:
        logging.debug(cmd)
        success = hoc.execute(cmd)
        if not success:
            raise HocError('Error produced by hoc command "%s"' % cmd)


def hoc_comment(comment):
    logging.debug(comment)


def psection(section):
    """
    function psection(section):

    Print info about section in a hoc format which is executable.
    (length, parent, diameter, membrane information)

    Use section.psection() instead to get a data structure that
    contains the same information and more.

    This function is deprecated and will be removed in a future
    release.

    See:

    https://www.neuron.yale.edu/neuron/static/py_doc/modelspec/programmatic/topology.html?#psection
    """
    warnings.warn(
        "neuron.psection() is deprecated; use print(sec.psection()) instead",
        DeprecationWarning,
        stacklevel=2,
    )
    h.psection(sec=section)


def init():
    """
    function init():

    Initialize the simulation kernel.  This should be called before a run(tstop) call.

    ** This function exists for historical purposes. Use in new code is not recommended. **

    Use h.finitialize() instead, which allows you to specify the membrane potential
    to initialize to; via e.g. h.finitialize(-65)

    This function is deprecated and will be removed in a future
    release.

    By default, the units used by h.finitialize are in mV, but you can be explicit using
    NEURON's unit's library, e.g.

    .. code-block:: python

        from neuron.units import mV
        h.finitialize(-65 * mV)

    https://www.neuron.yale.edu/neuron/static/py_doc/simctrl/programmatic.html?#finitialize

    """
    warnings.warn(
        "neuron.init() is deprecated; use h.init() instead",
        DeprecationWarning,
        stacklevel=2,
    )

    h.finitialize()


def run(tstop):
    """
    function run(tstop)

    Run the simulation (advance the solver) until tstop [ms]

    `h.run()` and `h.continuerun(tstop)` are more powerful solutions defined in the `stdrun.hoc` library.

    ** This function exists for historical purposes. Use in new code is not recommended. **

    This function is deprecated and will be removed in a future
    release.

    For running a simulation, consider doing the following instead:

    Begin your code with

    .. code-block:: python

        from neuron import h
        from neuron.units import ms, mV
        h.load_file('stdrun.hoc')

    Then when it is time to initialize and run the simulation:

    .. code-block:: python

        h.finitialize(-65 * mV)
        h.continuerun(100 * ms)

    where the initial membrane potential and the simulation run time are adjusted as appropriate
    for your model.

    """
    warnings.warn(
        "neuron.run(tstop) is deprecated; use h.stdinit() and h.continuerun(tstop) instead",
        DeprecationWarning,
        stacklevel=2,
    )

    h("tstop = %g" % tstop)
    h("while (t < tstop) { fadvance() }")
    # what about pc.psolve(tstop)?


_nrn_dll = None
_nrn_hocobj_ptr = None
_double_ptr = None
_double_size = None


def numpy_element_ref(numpy_array, index):
    """Return a HOC reference into a numpy array.

    Parameters
    ----------
    numpy_array : :class:`numpy.ndarray`
        the numpy array
    index : int
        the index into the numpy array

    .. warning::

        No bounds checking.

    .. warning::

        Assumes a contiguous array of doubles. In particular, be careful when
        using slices. If the array is multi-dimensional,
        the user must figure out the integer index to the desired element.
    """
    global _nrn_dll, _double_ptr, _double_size, _nrn_hocobj_ptr
    import ctypes

    if _nrn_hocobj_ptr is None:
        _nrn_hocobj_ptr = nrn_dll_sym("nrn_hocobj_ptr")
        _nrn_hocobj_ptr.restype = ctypes.py_object
        _double_ptr = ctypes.POINTER(ctypes.c_double)
        _double_size = ctypes.sizeof(ctypes.c_double)
    void_p = (
        ctypes.cast(numpy_array.ctypes.data_as(_double_ptr), ctypes.c_voidp).value
        + index * _double_size
    )
    return _nrn_hocobj_ptr(ctypes.cast(void_p, _double_ptr))


def nrn_dll_sym(name, type=None):
    """return the specified object from the NEURON dlls.

    Parameters
    ----------
    name : string
        the name of the object (function, integer, etc...)
    type : None or ctypes type (e.g. ctypes.c_int)
        the type of the object (if None, assumes function pointer)
    """
    # TODO: this won't work under Windows; will need to search through until
    #       can find the right dll (should we cache the results of the search?)
    import os

    if os.name == "nt":
        return nrn_dll_sym_nt(name, type)
    dll = nrn_dll()
    if type is None:
        return dll.__getattr__(name)
    else:
        return type.in_dll(dll, name)


nt_dlls = []


def nrn_dll_sym_nt(name, type):
    """return the specified object from the NEURON dlls.
    helper for nrn_dll_sym(name, type). Try to find the name in either
    nrniv.dll or libnrnpython1013.dll
    """
    global nt_dlls
    import ctypes
    import os

    if len(nt_dlls) == 0:
        b = "bin"
        if h.nrnversion(8).find("i686") == 0:
            b = "bin"
        path = os.path.join(h.neuronhome().replace("/", "\\"), b)
        fac = 10 if sys.version_info[1] < 10 else 100  # 3.9 is 39 ; 3.10 is 310
        p = sys.version_info[0] * fac + sys.version_info[1]
        for dllname in ["libnrniv.dll", "libnrnpython%d.dll" % p]:
            p = os.path.join(path, dllname)
            try:
                nt_dlls.append(ctypes.cdll[p])
            except:
                pass
    for dll in nt_dlls:
        try:
            a = dll.__getattr__(name)
        except:
            a = None
        if a:
            if type is None:
                return a
            else:
                return type.in_dll(dll, name)
    raise Exception("unable to connect to the NEURON library containing " + name)


def nrn_dll(printpath=False):
    """Return a ctypes object corresponding to the NEURON library.

    .. warning::

        This provides access to the C-language internals of NEURON and should
        be used with care.
    """
    import ctypes
    import glob
    import os
    import sys

    try:
        # extended? if there is a __file__, then use that
        if printpath:
            print("hoc.__file__ %s" % _original_hoc_file)
        the_dll = ctypes.pydll[_original_hoc_file]
        return the_dll
    except:
        pass

    success = False
    if sys.platform == "msys" or sys.platform == "win32":
        p = "hoc%d%d" % (sys.version_info[0], sys.version_info[1])
    else:
        p = "hoc"

    try:
        # maybe hoc.so in this neuron module
        base_path = os.path.join(os.path.split(__file__)[0], p)
        dlls = glob.glob(base_path + "*.*")
        for dll in dlls:
            try:
                the_dll = ctypes.pydll[dll]
                if printpath:
                    print(dll)
                return the_dll
            except:
                pass
    except:
        pass
    # maybe old default module location
    neuron_home = os.path.split(os.path.split(h.neuronhome())[0])[0]
    base_path = os.path.join(neuron_home, "lib", "python", "neuron", p)
    for extension in ["", ".dll", ".so", ".dylib"]:
        dlls = glob.glob(base_path + "*" + extension)
        for dll in dlls:
            try:
                the_dll = ctypes.pydll[dll]
                if printpath:
                    print(dll)
                success = True
            except:
                pass
            if success:
                break
        if success:
            break
    else:
        raise Exception("unable to connect to the NEURON library")
    return the_dll


def _modelview_mechanism_docstrings(dmech, tree):
    if dmech.name not in ("Ra", "capacitance"):
        docs = getattr(h, dmech.name).__doc__
        if docs.strip():
            for line in docs.split("\n"):
                tree.append(line, dmech.location, 0)


# TODO: put this someplace else
#       can't be in rxd because that would break things if no scipy
_sec_db = {}


def _declare_contour(secobj, obj, name):
    array, i = _parse_import3d_name(name)
    if obj is None:
        sec = getattr(h, array)[i]
    else:
        sec = getattr(obj, array)[i]
    j = secobj.first
    center_vec = secobj.contourcenter(
        secobj.raw.getrow(0), secobj.raw.getrow(1), secobj.raw.getrow(2)
    )
    x0, y0, z0 = [center_vec.x[i] for i in range(3)]
    # store a couple of points to check if the section has been moved
    pts = [(sec.x3d(i), sec.y3d(i), sec.z3d(i)) for i in [0, sec.n3d() - 1]]
    # (is_stack, x, y, z, xcenter, ycenter, zcenter)
    _sec_db[sec.hoc_internal_name()] = (
        True if secobj.contour_list else False,
        secobj.raw.getrow(0).c(j),
        secobj.raw.getrow(1).c(j),
        secobj.raw.getrow(2).c(j),
        x0,
        y0,
        z0,
        pts,
    )


def _create_all_list(obj):
    # used by import3d
    obj.all = []


def _create_sections_in_obj(obj, name, numsecs):
    # used by import3d to instantiate inside of a Python object
    setattr(
        obj,
        name,
        [h.Section(name="%s[%d]" % (name, i), cell=obj) for i in range(int(numsecs))],
    )


def _connect_sections_in_obj(obj, childsecname, childx, parentsecname, parentx):
    # used by import3d
    childarray, childi = _parse_import3d_name(childsecname)
    parentarray, parenti = _parse_import3d_name(parentsecname)
    getattr(obj, childarray)[childi].connect(
        getattr(obj, parentarray)[parenti](parentx), childx
    )


def _parse_import3d_name(name):
    if "[" in name:
        import re

        array, i = re.search(r"(.*)\[(\d*)\]", name).groups()
        i = int(i)
    else:
        array = name
        i = 0
    return array, i


def _pt3dstyle_in_obj(obj, name, x, y, z):
    # used by import3d
    array, i = _parse_import3d_name(name)
    h.pt3dstyle(1, x, y, z, sec=getattr(obj, array)[i])


def _pt3dadd_in_obj(obj, name, x, y, z, d):
    array, i = _parse_import3d_name(name)
    h.pt3dadd(x, y, z, d, sec=getattr(obj, array)[i])


def numpy_from_pointer(cpointer, size):
    buf_from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    buf_from_mem.restype = ctypes.py_object
    buf_from_mem.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
    cbuffer = buf_from_mem(cpointer, size * numpy.dtype(float).itemsize, 0x200)
    return numpy.ndarray((size,), float, cbuffer, order="C")


try:
    import ctypes
    import numpy
    import traceback

    vec_to_numpy_prototype = ctypes.CFUNCTYPE(
        ctypes.py_object, ctypes.c_int, ctypes.POINTER(ctypes.c_double)
    )

    def vec2numpy(size, data):
        try:
            return numpy_from_pointer(data, size)
        except:
            traceback.print_exc()
            return None

    vec_to_numpy_callback = vec_to_numpy_prototype(vec2numpy)
    set_vec_as_numpy = nrn_dll_sym("nrnpy_set_vec_as_numpy")
    set_vec_as_numpy(vec_to_numpy_callback)
except:
    pass


class _WrapperPlot:
    def __init__(self, data):
        """do not call directly"""
        self._data = data

    def __repr__(self):
        return "{}.plot()".format(repr(self._data))


class _RangeVarPlot(_WrapperPlot):
    """Plots the current state of the RangeVarPlot on the graph.

    Additional arguments and keyword arguments are passed to the graph's
    plotting method.

    Example, showing plotting to NEURON graphics, bokeh, matplotlib,
    plotnine/ggplot, and plotly:

    .. code::

      from matplotlib import pyplot
      from neuron import h, gui
      import bokeh.plotting as b
      import plotly
      import plotly.graph_objects as go
      import plotnine as p9
      import math

      dend = h.Section(name='dend')
      dend.nseg = 55
      dend.L = 6.28

      # looping over dend.allseg instead of dend to set 0 and 1 ends
      for seg in dend.allseg():
          seg.v = math.sin(dend.L * seg.x)

      r = h.RangeVarPlot('v', dend(0), dend(1))

      # matplotlib
      graph = pyplot.gca()
      r.plot(graph, linewidth=10, color='r')

      # NEURON Interviews graph
      g = h.Graph()
      r.plot(g, 2, 3)
      g.exec_menu('View = plot')

      # Bokeh
      bg = b.Figure()
      r.plot(bg, line_width=10)
      b.show(bg)

      # plotly
      r.plot(plotly).show()

      # also plotly
      fig = go.Figure()
      r.plot(fig)
      fig.show()

      pyplot.show()

      # plotnine/ggplot
      p9.ggplot() + r.plot(p9)

      # alternative plotnine/ggplot
      r.plot(p9.ggplot())
    """

    def __call__(self, graph, *args, **kwargs):
        yvec = h.Vector()
        xvec = h.Vector()
        self._data.to_vector(yvec, xvec)
        if isinstance(graph, hoc.HocObject):
            return yvec.line(graph, xvec, *args)
        str_type_graph = str(type(graph))
        if str_type_graph == "<class 'plotly.graph_objs._figure.Figure'>":
            # plotly figure
            import plotly.graph_objects as go

            kwargs.setdefault("mode", "lines")
            return graph.add_trace(go.Scatter(x=xvec, y=yvec, *args, **kwargs))
        if str_type_graph == "<class 'plotnine.ggplot.ggplot'>":
            # ggplot object
            import plotnine as p9
            import pandas as pd

            return graph + p9.geom_line(
                *args,
                data=pd.DataFrame({"x": xvec, "y": yvec}),
                mapping=p9.aes(x="x", y="y"),
                **kwargs,
            )
        str_graph = str(graph)
        if str_graph.startswith("<module 'plotly' from "):
            # plotly module
            import plotly.graph_objects as go

            fig = go.Figure()
            kwargs.setdefault("mode", "lines")
            return fig.add_trace(go.Scatter(x=xvec, y=yvec, *args, **kwargs))
        if str_graph.startswith("<module 'plotnine' from "):
            # plotnine module (contains ggplot)
            import plotnine as p9
            import pandas as pd

            return p9.geom_line(
                *args,
                data=pd.DataFrame({"x": xvec, "y": yvec}),
                mapping=p9.aes(x="x", y="y"),
                **kwargs,
            )
        if hasattr(graph, "plot"):
            # works with e.g. pyplot or a matplotlib axis
            return graph.plot(xvec, yvec, *args, **kwargs)
        if hasattr(graph, "line"):
            # works with e.g. bokeh
            return graph.line(xvec, yvec, *args, **kwargs)
        if str_type_graph == "<class 'matplotlib.figure.Figure'>":
            raise Exception("plot to a matplotlib axis not a matplotlib figure")
        raise Exception("Unable to plot to graphs of type {}".format(type(graph)))


class _PlotShapePlot(_WrapperPlot):
    """Plots the currently selected data on an object.

    Currently only pyplot is supported, e.g.

    from matplotlib import pyplot
    ps = h.PlotShape(False)
    ps.variable('v')
    ps.plot(pyplot)
    pyplot.show()

    Limitations: many. Currently only supports plotting a full cell colored based on a variable."""

    # TODO: handle pointmark, specified sections, color
    def __call__(self, graph, *args, **kwargs):
        from neuron.gui2.utilities import _segment_3d_pts

        def _get_pyplot_axis3d(fig):
            """requires matplotlib"""
            from matplotlib.pyplot import cm
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            import numpy as np

            class Axis3DWithNEURON(Axes3D):
                def auto_aspect(self):
                    """sets the x, y, and z range symmetric around the center

                    Probably needs a square figure to preserve lengths as you rotate."""
                    bounds = [self.get_xlim(), self.get_ylim(), self.get_zlim()]
                    half_delta_max = max([(item[1] - item[0]) / 2 for item in bounds])
                    xmid = sum(bounds[0]) / 2
                    ymid = sum(bounds[1]) / 2
                    zmid = sum(bounds[2]) / 2
                    self.auto_scale_xyz(
                        [xmid - half_delta_max, xmid + half_delta_max],
                        [ymid - half_delta_max, ymid + half_delta_max],
                        [zmid - half_delta_max, zmid + half_delta_max],
                    )

                def mark(self, segment, marker="or", **kwargs):
                    """plot a marker on a segment

                    Args:
                        segment = the segment to mark
                        marker = matplotlib marker
                        **kwargs = passed to matplotlib's plot
                    """
                    x, y, z = _get_3d_pt(segment)
                    self.plot([x], [y], [z], marker)
                    return self

                def _do_plot(
                    self, val_min, val_max, sections, variable, cmap=cm.cool, **kwargs
                ):
                    """
                    Plots a 3D shapeplot
                    Args:
                        sections = list of h.Section() objects to be plotted
                        **kwargs passes on to matplotlib (e.g. linewidth=2 for thick lines)
                    Returns:
                        lines = list of line objects making up shapeplot
                    """
                    # Adapted from
                    # https://github.com/ahwillia/PyNeuron-Toolbox/blob/master/PyNeuronToolbox/morphology.py
                    # Accessed 2019-04-11, which had an MIT license

                    # Default is to plot all sections.
                    if sections is None:
                        sections = list(h.allsec())

                    h.define_shape()

                    # default color is black
                    kwargs.setdefault("color", "black")

                    # Plot each segement as a line
                    lines = {}
                    lines_list = []
                    vals = []
                    for sec in sections:
                        all_seg_pts = _segment_3d_pts(sec)
                        for seg, (xs, ys, zs, _, _) in zip(sec, all_seg_pts):
                            (line,) = self.plot(xs, ys, zs, "-", **kwargs)
                            if variable is not None:
                                val = _get_variable_seg(seg, variable)
                                vals.append(val)
                                if val is not None:
                                    lines[line] = "%s at %s" % (val, seg)
                                else:
                                    lines[line] = str(seg)
                            else:
                                lines[line] = str(seg)
                            lines_list.append(line)
                    if variable is not None:
                        val_range = val_max - val_min
                        if val_range:
                            for sec in sections:
                                for line, val in zip(lines_list, vals):
                                    if val is not None:
                                        col = _get_color(
                                            variable,
                                            val,
                                            cmap,
                                            val_min,
                                            val_max,
                                            val_range,
                                        )
                                        line.set_color(col)
                    return lines

            return Axis3DWithNEURON(fig)

        def _get_variable_seg(seg, variable):
            if isinstance(variable, str):
                try:
                    if "." in variable:
                        mech, var = variable.split(".")
                        val = getattr(getattr(seg, mech), var)
                    else:
                        val = getattr(seg, variable)
                except AttributeError:
                    # leave default color if no variable found
                    val = None
            else:
                try:
                    vals = variable.nodes(seg).concentration
                    val = sum(vals) / len(vals)
                except:
                    val = None

            return val

        def _get_3d_pt(segment):
            import numpy as np

            # TODO: there has to be a better way to do this
            sec = segment.sec
            n3d = sec.n3d()
            arc3d = [sec.arc3d(i) for i in range(n3d)]
            x3d = np.array([sec.x3d(i) for i in range(n3d)])
            y3d = np.array([sec.y3d(i) for i in range(n3d)])
            z3d = np.array([sec.z3d(i) for i in range(n3d)])
            seg_l = sec.L * segment.x
            x = np.interp(seg_l, arc3d, x3d)
            y = np.interp(seg_l, arc3d, y3d)
            z = np.interp(seg_l, arc3d, z3d)
            return x, y, z

        def _do_plot_on_matplotlib_figure(fig):
            import ctypes

            get_plotshape_data = nrn_dll_sym("get_plotshape_data")
            get_plotshape_data.restype = ctypes.py_object
            variable, varobj, lo, hi, secs = get_plotshape_data(
                ctypes.py_object(self._data)
            )
            if varobj is not None:
                variable = varobj
            kwargs.setdefault("picker", 2)
            result = _get_pyplot_axis3d(fig)
            _lines = result._do_plot(lo, hi, secs, variable, *args, **kwargs)
            result._mouseover_text = ""

            def _onpick(event):
                if event.artist in _lines:
                    result._mouseover_text = _lines[event.artist]
                else:
                    result._mouseover_text = ""
                return True

            result.auto_aspect()
            fig.canvas.mpl_connect("pick_event", _onpick)

            def format_coord(*args):
                return result._mouseover_text

            result.format_coord = format_coord
            return result

        def _get_color(variable, val, cmap, lo, hi, val_range):
            if variable is None or val is None:
                col = "black"
            elif val_range == 0:
                if val < lo:
                    col = color_to_hex(cmap(0))
                elif val > hi:
                    col = color_to_hex(cmap(255))
                else:
                    val = color_to_hex(128)
            else:
                col = color_to_hex(
                    cmap(int(255 * (min(max(val, lo), hi) - lo) / (val_range)))
                )
            return col

        def color_to_hex(col):
            items = [hex(int(255 * col_item))[2:] for col_item in col][:-1]
            return "#" + "".join(
                [item if len(item) == 2 else "0" + item for item in items]
            )

        def _do_plot_on_plotly():
            """requires matplotlib for colormaps if not specified explicitly"""
            import ctypes
            import plotly.graph_objects as go

            class FigureWidgetWithNEURON(go.FigureWidget):
                def mark(self, segment, marker="or", **kwargs):
                    """plot a marker on a segment

                    Args:
                        segment = the segment to mark
                        **kwargs = passed to go.Scatter3D plot
                    """
                    x, y, z = _get_3d_pt(segment)
                    # approximately match the appearance of the matplotlib defaults
                    kwargs.setdefault("marker_size", 5)
                    kwargs.setdefault("marker_color", "red")
                    kwargs.setdefault("marker_opacity", 1)

                    self.add_trace(
                        go.Scatter3d(
                            x=[x],
                            y=[y],
                            z=[z],
                            name="",
                            hovertemplate=str(segment),
                            **kwargs,
                        )
                    )
                    return self

            get_plotshape_data = nrn_dll_sym("get_plotshape_data")
            get_plotshape_data.restype = ctypes.py_object
            variable, varobj, lo, hi, secs = get_plotshape_data(
                ctypes.py_object(self._data)
            )
            if varobj is not None:
                variable = varobj
            if secs is None:
                secs = list(h.allsec())

            if variable is None:
                kwargs.setdefault("color", "black")

                data = []
                for sec in secs:
                    xs = [sec.x3d(i) for i in range(sec.n3d())]
                    ys = [sec.y3d(i) for i in range(sec.n3d())]
                    zs = [sec.z3d(i) for i in range(sec.n3d())]
                    data.append(
                        go.Scatter3d(
                            x=xs,
                            y=ys,
                            z=zs,
                            name="",
                            hovertemplate=str(sec),
                            mode="lines",
                            line=go.scatter3d.Line(color=kwargs["color"], width=2),
                        )
                    )
                return FigureWidgetWithNEURON(data=data, layout={"showlegend": False})

            else:
                if "cmap" not in kwargs:
                    # use same default colormap as the matplotlib version
                    from matplotlib.pyplot import cm

                    kwargs["cmap"] = cm.cool

                cmap = kwargs["cmap"]
                show_diam = False

                # calculate bounds

                val_range = hi - lo

                data = []
                for sec in secs:
                    all_seg_pts = _segment_3d_pts(sec)
                    for seg, (xs, ys, zs, _, _) in zip(sec, all_seg_pts):
                        val = _get_variable_seg(seg, variable)
                        hover_template = str(seg)
                        if val is not None:
                            hover_template += "<br>" + ("%.3f" % val)
                        col = _get_color(variable, val, cmap, lo, hi, val_range)
                        if show_diam:
                            diam = seg.diam
                        else:
                            diam = 2
                        data.append(
                            go.Scatter3d(
                                x=xs,
                                y=ys,
                                z=zs,
                                name="",
                                hovertemplate=hover_template,
                                mode="lines",
                                line=go.scatter3d.Line(color=col, width=diam),
                            )
                        )

                return FigureWidgetWithNEURON(data=data, layout={"showlegend": False})

        if hasattr(graph, "__name__"):
            if graph.__name__ == "matplotlib.pyplot":
                fig = graph.figure()
                return _do_plot_on_matplotlib_figure(fig)
            elif graph.__name__ == "plotly":
                return _do_plot_on_plotly()
        elif str(type(graph)) == "<class 'matplotlib.figure.Figure'>":
            return _do_plot_on_matplotlib_figure(graph)
        raise NotImplementedError


def _nmodl():
    try:
        import nmodl.dsl as nmodl

        return nmodl
    except ModuleNotFoundError:
        raise Exception(
            "Missing nmodl module; install from https://github.com/bluebrain/nmodl"
        )


class DensityMechanism:
    def __init__(self, name):
        """Initialize the DensityMechanism.

        Takes the name of a range mechanism; call via e.g. neuron.DensityMechanism('hh')
        """
        self.__name = name
        self.__mt = h.MechanismType(0)
        self.__mt.select(-1)
        self.__mt.select(name)
        if self.__mt.selected() == -1:
            raise Exception("No DensityMechanism: " + name)
        self.__has_nmodl = False
        self.__ast = None
        self.__ions = None
        try:
            import nmodl

            self.__has_nmodl = True
        except ModuleNotFoundError:
            pass

    def __repr__(self):
        return "neuron.DensityMechanism(%r)" % self.__name

    def __dir__(self):
        my_dir = ["code", "file", "insert", "uninsert", "__repr__", "__str__"]
        if self.__has_nmodl:
            my_dir += ["ast", "ions", "ontology_ids"]
        return sorted(my_dir)

    @property
    def ast(self):
        """Abstract Syntax Tree representation.

        Requires the nmodl module, available from: https://github.com/bluebrain/nmodl

        The model is parsed on first access, and the results are cached for quick reaccess
        using the same neuron.DensityMechanism instance.
        """
        if self.__ast is None:
            nmodl = _nmodl()
            driver = nmodl.NmodlDriver()
            self.__ast = driver.parse_string(self.code)
        return self.__ast

    @property
    def code(self):
        """source code"""
        return self.__mt.code()

    @property
    def file(self):
        """source file path"""
        return self.__mt.file()

    def insert(self, secs):
        """insert this mechanism into a section or iterable of sections"""
        if isinstance(secs, nrn.Section):
            secs = [secs]
        for sec in secs:
            sec.insert(self.__name)

    def uninsert(self, secs):
        """uninsert (remove) this mechanism from a section or iterable of sections"""
        if isinstance(secs, nrn.Section):
            secs = [secs]
        for sec in secs:
            sec.uninsert(self.__name)

    @property
    def ions(self):
        """Dictionary of the ions involved in this mechanism"""
        if self.__ions is None:
            nmodl = _nmodl()
            lookup_visitor = nmodl.visitor.AstLookupVisitor()
            ions = lookup_visitor.lookup(self.ast, nmodl.ast.AstNodeType.USEION)
            result = {}
            for ion in ions:
                name = nmodl.to_nmodl(ion.name)
                read = [nmodl.to_nmodl(item) for item in ion.readlist]
                write = [nmodl.to_nmodl(item) for item in ion.writelist]
                if ion.valence:
                    valence = int(nmodl.to_nmodl(ion.valence.value))
                else:
                    valence = None
                ontology_id = None
                try:
                    ontology_id = ion.ontology_id
                except:
                    # older versions of the NMODL library didn't support .ontology_id
                    pass
                result[name] = {
                    "read": read,
                    "write": write,
                    "charge": valence,
                    "ontology_id": nmodl.to_nmodl(ontology_id),
                }
            self.__ions = result
        # return a copy
        return dict(self.__ions)

    @property
    def ontology_ids(self):
        nmodl = _nmodl()
        lookup_visitor = nmodl.visitor.AstLookupVisitor()

        try:
            onts = lookup_visitor.lookup(
                self.ast, nmodl.ast.AstNodeType.ONTOLOGY_STATEMENT
            )
        except AttributeError:
            raise Exception(
                "nmodl module out of date; missing support for ontology declarations"
            )

        return [nmodl.to_nmodl(ont.ontology_id) for ont in onts]


_store_savestates = []
_restore_savestates = []
_id_savestates = []


def register_savestate(id_, store, restore):
    """register routines to be called during SaveState

    id_ -- unique id (consider using a UUID)
    store -- called when saving the state to the object; returns a bytestring
    restore -- called when loading the state from the object; receives a bytestring
    """
    _id_savestates.append(id_)
    _store_savestates.append(store)
    _restore_savestates.append(restore)


def _store_savestate():
    import array
    import itertools

    version = 0
    result = [array.array("Q", [version]).tobytes()]
    for id_, store in zip(_id_savestates, _store_savestates):
        data = store()
        if len(data):
            result.append(
                array.array("Q", [len(id_)]).tobytes()
                + bytes(id_.encode("utf8"))
                + array.array("Q", [len(data)]).tobytes()
                + data
            )
    if len(result) == 1:
        # if no data to save, then don't even bother with a version
        result = []
    return bytearray(itertools.chain.from_iterable(result))


def _restore_savestate(data):
    import array

    # convert from bytearray
    data = bytes(data)
    metadata = array.array("Q")
    metadata.frombytes(data[:8])
    version = metadata[0]
    if version != 0:
        raise Exception("Unsupported SaveState version")
    position = 8
    while position < len(data):
        metadata = array.array("Q")
        metadata.frombytes(data[position : position + 8])
        name_length = metadata[0]
        position += 8
        name = data[position : position + name_length].decode("utf8")
        position += name_length
        metadata = array.array("Q")
        metadata.frombytes(data[position : position + 8])
        data_length = metadata[0]
        position += 8
        my_data = data[position : position + data_length]
        position += data_length
        # lookup the index because not everything that is registered is used
        try:
            index = _id_savestates.index(name)
        except ValueError:
            raise Exception("Undefined SaveState type " + name)
        _restore_savestates[index](my_data)
    if position != len(data):
        raise Exception("SaveState length error")


try:
    import ctypes

    def _rvp_plot(rvp):
        return _RangeVarPlot(rvp)

    def _plotshape_plot(ps):
        h.define_shape()
        return _PlotShapePlot(ps)

    _mech_classes = {}

    def _get_mech_object(name):
        if name in _mech_classes:
            my_class = _mech_classes[name]
        else:
            code = DensityMechanism(name).code
            # docstring is the title and a leading comment, if any
            inside_comment = False
            title = ""
            comment = []
            for line in code.split("\n"):
                line_s = line.strip()
                lower = line_s.lower()
                if inside_comment:
                    if lower.startswith("endcomment"):
                        break
                    comment.append(line)
                elif lower.startswith("title "):
                    title = line_s[6:]
                elif lower.startswith("comment"):
                    inside_comment = True
                elif line_s:
                    break

            docstring = title + "\n\n"

            docstring += "\n".join(comment)
            docstring = docstring.strip()

            clsdict = {"__doc__": docstring, "title": title}
            my_class = type(name, (DensityMechanism,), clsdict)
            _mech_classes[name] = my_class
        return my_class(name)

    set_toplevel_callbacks = nrn_dll_sym("nrnpy_set_toplevel_callbacks")
    _rvp_plot_callback = ctypes.py_object(_rvp_plot)
    _plotshape_plot_callback = ctypes.py_object(_plotshape_plot)
    _get_mech_object_callback = ctypes.py_object(_get_mech_object)
    _restore_savestate_callback = ctypes.py_object(_restore_savestate)
    _store_savestate_callback = ctypes.py_object(_store_savestate)
    set_toplevel_callbacks(
        _rvp_plot_callback,
        _plotshape_plot_callback,
        _get_mech_object_callback,
        _store_savestate_callback,
        _restore_savestate_callback,
    )
except:
    pass


def _has_scipy():
    """
    to check for scipy:

    has_scipy = 0
    objref p
    if (nrnpython("import neuron")) {
        p = new PythonObject()
        has_scipy = p.neuron._has_scipy()
    }
    """
    try:
        import scipy
    except:
        return 0
    return 1


def _pkl(arg):
    # print 'neuron._pkl arg is ', arg
    return h.Vector(0)


def format_exception(type, value, tb):
    """Single string return wrapper for traceback.format_exception
    used by nrnpyerr_str
    """
    import traceback

    slist = (
        traceback.format_exception_only(type, value)
        if tb is None
        else traceback.format_exception(type, value, tb)
    )
    s = "".join(slist)
    return s


def nrnpy_pass():
    return 1


def nrnpy_pr(stdoe, s):
    if stdoe == 1:
        sys.stdout.write(s.decode())
    else:
        sys.stderr.write(s.decode())
        sys.stderr.flush()
    return 0


if not embedded:
    try:
        # nrnpy_pr callback in place of hoc printf
        # ensures consistent with python stdout even with jupyter notebook.
        # nrnpy_pass callback used by h.doNotify() in MINGW when not called from
        # gui thread in order to allow the gui thread to run.

        nrnpy_set_pr_etal = nrn_dll_sym("nrnpy_set_pr_etal")

        nrnpy_pr_proto = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.c_char_p)
        nrnpy_pass_proto = ctypes.CFUNCTYPE(ctypes.c_int)
        nrnpy_set_pr_etal.argtypes = [nrnpy_pr_proto, nrnpy_pass_proto]

        nrnpy_pr_callback = nrnpy_pr_proto(nrnpy_pr)
        nrnpy_pass_callback = nrnpy_pass_proto(nrnpy_pass)
        nrnpy_set_pr_etal(nrnpy_pr_callback, nrnpy_pass_callback)
    except:
        print("Failed to setup nrnpy_pr")
        pass


def nrnpy_vec_math(op, flag, arg1, arg2=None):
    import numbers

    valid_types = (numbers.Number, hoc.HocObject)
    if isinstance(arg1, valid_types):
        if flag == 2:
            # unary
            arg1 = arg1.c()
            if op == "uneg":
                return arg1.mul(-1)
            if op == "upos":
                return arg1
            if op == "uabs":
                return arg1.abs()
        elif isinstance(arg2, valid_types):
            if flag == 1:
                # either reversed (flag=1) or unary (flag=2)
                arg2 = arg2.c()
                if op in ("mul", "add"):
                    return getattr(arg2, op)(arg1)
                if op == "div":
                    return arg2.pow(-1).mul(arg1)
                if op == "sub":
                    return arg2.mul(-1).add(arg1)
            else:
                arg1 = arg1.c()
                return getattr(arg1, op)(arg2)

    return NotImplemented


def _nrnpy_rvp_pyobj_callback(f):
    # unless f is an rxd variable, we return it directly
    f_type = str(type(f))
    if f_type not in (
        "<class 'neuron.rxd.species.SpeciesOnRegion'>",
        "<class 'neuron.rxd.species.Species'>",
    ):
        return f

    # if we're here, f is an rxd variable, and we return a function that looks
    # up the weighted average concentration given an x and h.cas()
    # this is not particularly efficient so it is probably better to use this for
    # fixed timepoints rather than displays that update mid-simulation
    fref = weakref.ref(f)

    def result(x):
        sp = fref()
        if sp:
            try:
                # h.cas() will fail if there are no sections
                nodes = sp.nodes(h.cas()(x))
            except:
                return None
            if nodes:
                total_volume = sum(node.volume for node in nodes)
                return (
                    sum(node.concentration * node.volume for node in nodes)
                    / total_volume
                )
            else:
                return None
        return None

    return result


try:
    nrnpy_vec_math_register = nrn_dll_sym("nrnpy_vec_math_register")
    nrnpy_vec_math_register(ctypes.py_object(nrnpy_vec_math))
except:
    print("Failed to setup nrnpy_vec_math")

try:
    _nrnpy_rvp_pyobj_callback_register = nrn_dll_sym(
        "nrnpy_rvp_pyobj_callback_register"
    )
    _nrnpy_rvp_pyobj_callback_register(ctypes.py_object(_nrnpy_rvp_pyobj_callback))
except:
    print("Failed to setup _nrnpy_rvp_pyobj_callback")

try:
    from neuron.psection import psection

    nrn.set_psection(psection)
except:
    print("Failed to setup nrn.Section.psection")
pass

import atexit as _atexit


@_atexit.register
def clear_gui_callback():
    try:
        nrnpy_set_gui_callback = nrn_dll_sym("nrnpy_set_gui_callback")
        nrnpy_set_gui_callback(None)
    except:
        pass


try:
    from IPython import get_ipython as _get_ipython
except:
    _get_ipython = lambda *args: None


def _hocobj_html(item):
    try:
        if item.hname().split("[")[0] == "ModelView":
            return _mview_html_tree(item.display.top)
        return None
    except:
        return None


def _mview_html_tree(hlist, inside_mechanisms_in_use=0):
    items = []
    if inside_mechanisms_in_use:
        miu_level = inside_mechanisms_in_use + 1
    else:
        miu_level = 0
    my_miu_level = miu_level
    for ho in hlist:
        html = ho.s.lstrip(" *")
        if ho.children:
            if html == "Mechanisms in use":
                my_miu_level = 1
        if html or miu_level == 3:
            if ho.children:
                children_data = _mview_html_tree(
                    ho.children, inside_mechanisms_in_use=my_miu_level
                )
                if miu_level == 3:
                    items.append(html + children_data)
                else:
                    items.append(
                        f"<div><details><summary style='cursor:pointer'>{html}</summary><div style='margin-left:1.06em'>{children_data}</div></div>"
                    )
            else:
                if miu_level == 3:
                    items.append(html)
                else:
                    items.append(f"<div style='margin-left:1.06em'>{html}</div>")

    if miu_level == 3:
        return f"{'<br>'.join(items)}"
    else:
        return f"{''.join(items)}"


# register our ModelView display formatter with Jupyter if available
if _get_ipython() is not None:
    html_formatter = _get_ipython().display_formatter.formatters["text/html"]
    html_formatter.for_type(hoc.HocObject, _hocobj_html)



from neuron import h
from ballandstick import BallAndStick

### MPI must be initialized before we create a ParallelContext object
h.nrnmpi_init()
pc = h.ParallelContext()


class Ring:
    """A network of *N* ball-and-stick cells where cell n makes an
    excitatory synapse onto cell n + 1 and the last, Nth cell in the
    network projects to the first cell.
    """

    def __init__(
        self, N=5, stim_w=0.04, stim_t=9, stim_delay=1, syn_w=0.01, syn_delay=5, r=50
    ):
        """
        :param N: Number of cells.
        :param stim_w: Weight of the stimulus
        :param stim_t: time of the stimulus (in ms)
        :param stim_delay: delay of the stimulus (in ms)
        :param syn_w: Synaptic weight
        :param syn_delay: Delay of the synapse
        :param r: radius of the network
        """
        self._N = N
        self.set_gids()  ### assign gids to processors
        self._syn_w = syn_w
        self._syn_delay = syn_delay
        self._create_cells(r)  ### changed to use self._N instead of passing in N
        self._connect_cells()
        ### the 0th cell only exists on one process... that's the only one that gets a netstim
        if pc.gid_exists(0):
            self._netstim = h.NetStim()
            self._netstim.number = 1
            self._netstim.start = stim_t
            self._nc = h.NetCon(
                self._netstim, pc.gid2cell(0).syn
            )  ### grab cell with gid==0 wherever it exists
            self._nc.delay = stim_delay
            self._nc.weight[0] = stim_w

    def set_gids(self):
        """Set the gidlist on this host."""
        #### Round-robin counting.
        #### Each host has an id from 0 to pc.nhost() - 1.
        self.gidlist = list(range(pc.id(), self._N, pc.nhost()))
        for gid in self.gidlist:
            pc.set_gid2node(gid, pc.id())

    def _create_cells(self, r):
        self.cells = []
        for i in self.gidlist:  ### only create the cells that exist on this host
            theta = i * 2 * h.PI / self._N
            self.cells.append(
                BallAndStick(i, h.cos(theta) * r, h.sin(theta) * r, 0, theta)
            )
        ### associate the cell with this host and gid
        for cell in self.cells:
            pc.cell(cell._gid, cell._spike_detector)

    def _connect_cells(self):
        ### this method is different because we now must use ids instead of objects
        for target in self.cells:
            source_gid = (target._gid - 1 + self._N) % self._N
            nc = pc.gid_connect(source_gid, target.syn)
            nc.weight[0] = self._syn_w
            nc.delay = self._syn_delay
            target._ncs.append(nc)



from mpi4py import MPI
from neuron import h
import sys

pc = h.ParallelContext()

s = "mpi4py thinks I am %d of %d\
\nNEURON thinks I am %d of %d"

cw = MPI.COMM_WORLD
print(s % (cw.rank, cw.size, pc.id(), pc.nhost()))

pc.done()

# Desired behaviour:

## emuller@drishti ~/hg/nrn_neurens_hg/src/nrnpython/examples $ mpiexec -np 4 nrniv -mpi -python test0.py
## numprocs=4
## NEURON -- VERSION 7.0.unoffical-nrn-hg.unknown (2216) 2008-09-15
## Duke, Yale, and the BlueBrain Project -- Copyright 1984-2008
## See http://neuron.yale.edu/credits.html

## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## numprocs=4
## mpi4py thinks I am 0 of 4, NEURON thinks I am 0 of 4
## mpi4py thinks I am 2 of 4, NEURON thinks I am 2 of 4

## mpi4py thinks I am 1 of 4, NEURON thinks I am 1 of 4
## mpi4py thinks I am 3 of 4, NEURON thinks I am 3 of 4


## >>>
## emuller@drishti ~/hg/nrn_neurens_hg/src/nrnpython/examples $


## emuller@drishti ~/hg/nrn_neurens_hg/src/nrnpython/examples $ mpiexec -np 4 python test0.py
## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## MPI_Initialized==true, enabling MPI funtionality.
## numprocs=4
## NEURON -- VERSION 7.0.unoffical-nrn-hg.unknown (2216) 2008-09-15
## Duke, Yale, and the BlueBrain Project -- Copyright 1984-2008
## See http://neuron.yale.edu/credits.html

## mpi4py thinks I am 3 of 4, NEURON thinks I am 3 of 4

## mpi4py thinks I am 2 of 4, NEURON thinks I am 2 of 4

## mpi4py thinks I am 1 of 4, NEURON thinks I am 1 of 4

## mpi4py thinks I am 0 of 4, NEURON thinks I am 0 of 4

## emuller@drishti ~/hg/nrn_neurens_hg/src/nrnpython/examples $ python test0.py
## MPI_Initialized==true, enabling MPI funtionality.
## numprocs=1
## NEURON -- VERSION 7.0.unoffical-nrn-hg.unknown (2216) 2008-09-15
## Duke, Yale, and the BlueBrain Project -- Copyright 1984-2008
## See http://neuron.yale.edu/credits.html

## mpi4py thinks I am 0 of 1, NEURON thinks I am 0 of 1


## emuller@drishti ~/hg/nrn_neurens_hg/src/nrnpython/examples $ python
## Python 2.5.2 (r252:60911, Aug 22 2008, 18:45:05)
## [GCC 4.1.2 (Gentoo 4.1.2 p1.1)] on linux2
## Type "help", "copyright", "credits" or "license" for more information.
## >>> import neuron
## MPI_Initialized==false, disabling MPI funtionality.
## NEURON -- VERSION 7.0.unoffical-nrn-hg.unknown (2216) 2008-09-15
## Duke, Yale, and the BlueBrain Project -- Copyright 1984-2008
## See http://neuron.yale.edu/credits.html

## >>>



"""
Extensions of the nrn.Section class.

$Id:$
"""

import neuron
from neuron import nrn
import sys
import io


class ExtendedSection(nrn.Section):
    def __init__(self, name=None):
        nrn.Section.__init__(self)
        self.hoc_name = nrn.Section.name(self)
        if name:
            self._name = name
        else:
            self._name = self.hoc_name
        self.mechanisms = set()

    def __getattr__(self, name):
        if name in self.mechanisms:
            return AllSegments(self, name)
        else:
            return self.__getattribute__(name)

    def insert(self, mech_name):
        self.mechanisms.add(mech_name)
        return nrn.Section.insert(self, mech_name)

    def _get_name(self):
        return self._name

    def _set_name(self, name):
        assert isinstance(name, str)
        self._name = name

    name = property(fget=_get_name, fset=_set_name)

    def psection(self):
        if self.name != self.hoc_name:
            sys.stdout.write("%s:  " % self.name)
        neuron.psection(self)


class AllSegments(object):
    """Allows setting membrane properties for all segments in a section at once,
    without having to explicitly iterate over all segments.

    Example usage:
        >>> soma = neuron.sections.ExtendedSection()
        >>> soma.nseg = 3
        >>> soma.insert('pas')
        >>> type(soma.pas)
        <class 'neuron.sections.AllSegments'>
        >>> soma.pas.e = -72
        >>> soma.pas.e
        [-72.0, -72.0, -72.0]
        >>> soma(0.5).pas.e
        -72.0
    """

    def __init__(self, sec, mech_name):
        self.sec = sec
        self.mech_name = mech_name

    def __setattr__(self, name, value):
        success = False
        for seg in self.sec:
            if hasattr(seg, self.mech_name):
                mech = getattr(seg, self.mech_name)
                if hasattr(mech, name):
                    mech.name = value
                    success = True
        if not success:
            try:
                object.__setattr__(self, name, value)
            except AttributeError:
                raise AttributeError(
                    "%s mechanism has no attribute %s" % (self.mech_name, name)
                )

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            return [
                getattr(getattr(seg, self.mech_name), name)
                for seg in self.sec
                if hasattr(seg, self.mech_name)
            ]



# test_a_neuronoptions.py named to be first test file in this folder.
# This is so testing with pytest will do this first. Otherwise
# neuron will already be imported and this test will not be substantive
# as it relies on being the first code to 'import neuron'. If neuron is
# already imported, a warning will be raised.
# This test trivially increases NFRAME by a small amount and tests
# that increase. Should not affect other tests in this folder.
# When this test was written, default NSTACK=1000 and NFRAME=512
#


import sys, warnings

neuron_already_imported = "neuron" in sys.modules

import os

nrnoptions = "-nogui -NSTACK 3000 -NFRAME 525"
os.environ["NEURON_MODULE_OPTIONS"] = nrnoptions

from neuron import h

# test that HOC NFRAME is actually larger than the default 512
h(
    """
func add_recurse() {
  if ($1 == 0) {
    return 0
  }
  return $1 + add_recurse($1-1)
}
"""
)


def test_neuronoptions():
    nrnver7 = h.nrnversion(7)  # First token is NEURON and may have -dll option
    print(nrnoptions)
    print(nrnver7)
    if neuron_already_imported:
        warnings.warn(
            UserWarning("neuron was already imported prior to setting os.environ")
        )
    else:
        assert nrnoptions in nrnver7
        i = 520
        assert h.add_recurse(i) == i * (i + 1) / 2


if __name__ == "__main__":
    test_neuronoptions()



import os
import pytest
from testutils import compare_data, tol


@pytest.fixture
def cell_model(neuron_instance):
    """Provide a Cell class that use the simple.asc Neurolucida morphology file."""

    h, rxd, data, save_path = neuron_instance

    class Cell:
        """Create a model cell based on the simple.asc Neurolucida file.

        Args:
            shift (triplet, optional) displacement in  x, y, z of the cell,
                default (0,0,0).
        """

        def __init__(self, shift=(0, 0, 0)):
            self.cell = h.Import3d_Neurolucida3()
            path = os.path.dirname(os.path.abspath(__file__))
            self.cell.input(os.path.join(path, "simple.asc"))
            i3d = h.Import3d_GUI(self.cell, False)
            i3d.instantiate(self)
            for sec in self.all:
                sec.nseg = int(sec.L / 40) * 2 + 1
            sx, sy, sz = shift
            for sec in self.soma:
                pts = [
                    (sec.x3d(i), sec.y3d(i), sec.z3d(i), sec.diam3d(i))
                    for i in range(sec.n3d())
                ]
                sec.pt3dclear()
                for (x, y, z, diam) in pts:
                    sec.pt3dadd(x + sx, y + sy, z + sz, diam)

    yield (h, rxd, data, save_path, Cell)


def test_soma_outlines(cell_model):
    """Import a Neurolucida model and use the soma outline for voxelization."""

    h, rxd, data, save_path, Cell = cell_model
    c = Cell()
    rxd.set_solve_type(c.all, dimension=3)
    cyt = rxd.Region(c.all, name="cyt", nrn_region="i", dx=1.0)
    ca = rxd.Species(
        cyt, initial=lambda node: 1 if node in c.soma[0] else 0, d=0.5, charge=2
    )

    h.finitialize(-65)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_soma_move(cell_model):
    """Import a Neurolucida model and move it."""

    h, rxd, data, save_path, Cell = cell_model
    c = Cell(shift=(20, 10, 20))
    rxd.set_solve_type(c.all, dimension=3)
    cyt = rxd.Region(c.all, name="cyt", nrn_region="i", dx=1.0)
    ca = rxd.Species(
        cyt, initial=lambda node: 1 if node in c.soma[0] else 0, d=0.5, charge=2
    )

    h.finitialize(-65)
    h.continuerun(10)
    assert min([nd.x3d for nd in ca.nodes(c.soma[0])]) > 29
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_multiple_soma(cell_model):
    """Create two Neurolucida models."""

    h, rxd, data, save_path, Cell = cell_model
    c = Cell()
    c2 = Cell(shift=(20, 10, 20))
    rxd.set_solve_type(c.all, dimension=3)
    cyt = rxd.Region(c.all, name="cyt", nrn_region="i", dx=1.0)
    ca = rxd.Species(
        cyt,
        initial=lambda node: 1 if node in c.soma[0] or node in c2.soma[0] else 0,
        d=0.5,
        charge=2,
    )

    h.finitialize(-65)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_import_into_HOC_template(neuron_instance):
    """test that we can import at the top level without error"""
    h, rxd, data, save_path = neuron_instance
    h(
        """
        begintemplate HocTemplateTest
        objref this, somatic, all, axonal, basal
        proc init() {localobj nl, import
            all = new SectionList()
            somatic = new SectionList()
            axonal = new SectionList()
            basal = new SectionList()

            nl = new Import3d_Neurolucida3()
            nl.input("%s")
            import = new Import3d_GUI(nl, 0)
            import.instantiate(this)
        }

        create soma[1], dend[1]
        endtemplate HocTemplateTest"""
        % os.path.join(os.path.dirname(os.path.abspath(__file__)), "simple.asc")
    )
    cell = h.HocTemplateTest()
    assert len(list(cell.all)) == 2
    # the next two lines are a work-around for issue #1164
    cell = None
    sl = h.SectionList()


def test_toplevel_import(neuron_instance):
    """test that we can import at the top level without error"""
    h, rxd, data, save_path = neuron_instance
    cell = h.Import3d_Neurolucida3()
    path = os.path.dirname(os.path.abspath(__file__))
    cell.input(os.path.join(path, "simple.asc"))
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)
    assert len(list(h.allsec())) == 2
    for sec in h.allsec():
        h.delete_section(sec=sec)



from neuron import rxd
from neuron.rxd.rxdException import RxDException
from xml.etree import ElementTree as ET
import xml.dom.minidom


# species declaration with arguments [name, compartment, and initial amount]
class species:
    def __init__(self, name, compartment, initial_amount):
        self.name = name
        self.compartment = compartment
        self.initial_amount = str(initial_amount)


# compartment declaration with arguments [name, size]
class compartment:
    def __init__(self, name, size):
        self.name = name
        self.size = str(size)


class parameter:
    def __init__(self, name, value):
        self.name = name
        self.value = value


# reactants and products are of the form [[name,stoichiometry],...]
# reaction declaration that takes name, reversible (boolean), reactants, products, and an element tree
# representing the kinetic law
class reaction:
    def __init__(self, name, reversible, reactants, products, tree, modifiers):
        self.name = name
        self.reversible = reversible
        self.reactants = reactants
        self.modifiers = modifiers
        self.products = products
        self.kinetic_law = tree


# define units here
# units are of the form [kind, exponent, scale]
class unit_definition:
    def __init__(self, name):
        self.name = name
        self.units = list()


# structure defined as the middle man between NEURON and SBML file types
# it creates an smbl file for a single section or segment
class middle_man:
    # defining an empty structure to later add more things to list
    def __init__(self, name):
        self.name = name
        # dict of species in the simulation and their corresponding compartment, indexed by id
        # id = name + compartment
        self.species = dict()
        # dict of compartments
        # id = compartment + index of compartment
        self.compartments = dict()
        self.num_compartments = 0
        self.reactions = list()
        self.parameters = dict()
        # Of the form [kind,exponent,scale]
        self.unit_defs = dict()

    # The parts that are required for a species are name and compartment, and optionally initial amount
    def add_species(self, name, compartment, initial_amount=0):
        if type(name) != str or len(name) < 1:
            return 1
        elif type(compartment) != str or len(compartment) < 1:
            return 2

        temp_species = species(name, compartment, initial_amount)
        id_name = name + "_" + compartment
        self.species[id_name] = temp_species

    # add compartment to structure
    def add_compartment(self, name, size):
        if type(name) != str or len(name) < 1:
            return 1
        if size is None:
            size = 0.1
        temp_compartment = compartment(name, size)
        id_name = name + str(self.num_compartments)
        self.compartments[id_name] = temp_compartment
        self.num_compartments += 1
        return 0

    def add_parameter(self, name, value):
        self.parameters[name] = parameter(name, value)

    # add reaction to structure
    def add_reaction(
        self, name, reversible, reactants, products, kinetic_law, modifiers
    ):
        self.reactions.append(
            reaction(name, reversible, reactants, products, kinetic_law, modifiers)
        )

    # add a unit to be defined later
    def add_unit_def(self, name):
        temp_unitdef = unit_definition(name)
        self.unit_defs[name] = temp_unitdef

    # add a unit to an already defined unit
    def add_unit(self, unit_def, kind, exponent, scale):
        self.unit_defs[unit_def].units.append([kind, exponent, scale])

    # returns the SBML as an element tree
    def create_xml(self):
        sbml = ET.Element(
            "sbml",
            level="2",
            version="3",
            xmlns="http://www.sbml.org/sbml/level2/version3",
        )
        model = ET.SubElement(sbml, "model", name=self.name)
        if len(self.unit_defs) > 0:
            unitdefinitions = ET.SubElement(model, "listOfUnitDefinitions")
        for key, value in self.unit_defs.items():
            unitDefinition = ET.SubElement(
                unitdefinitions, "unitDefinition", id=value.name, name=value.name
            )
            listOfUnits = ET.SubElement(unitDefinition, "listOfUnits")
            for j in value.units:
                ET.SubElement(
                    listOfUnits, "unit", kind=j[0], exponent=str(j[1]), scale=str(j[2])
                )

        if len(self.compartments) > 0:
            listOfCompartments = ET.SubElement(model, "listOfCompartments")
        for key, value in self.compartments.items():
            ET.SubElement(
                listOfCompartments,
                "compartment",
                id=value.name,
                name=value.name,
                size=str(value.size),
            )

        if len(self.species) > 0:
            listOfSpecies = ET.SubElement(model, "listOfSpecies")
        for key, value in self.species.items():
            ET.SubElement(
                listOfSpecies,
                "species",
                compartment=value.compartment,
                id=key,
                name=value.name,
                initialConcentration=value.initial_amount,
            )

        if len(self.parameters) > 0:
            listOfParameters = ET.SubElement(model, "listOfParameters")
        for para, value in self.parameters.items():
            ET.SubElement(
                listOfParameters, "parameter", id=para, value=str(value.value)
            )

        if len(self.reactions) > 0:
            listOfReactions = ET.SubElement(model, "listOfReactions")
        for react in self.reactions:
            current_reaction = ET.SubElement(
                listOfReactions,
                "reaction",
                id=react.name,
                name=react.name,
                reversible=str(react.reversible).lower(),
            )
            if len(react.reactants) > 0:
                listOfReactants = ET.SubElement(current_reaction, "listOfReactants")
            for i in react.reactants:
                ET.SubElement(
                    listOfReactants,
                    "speciesReference",
                    species=i[0],
                    stoichiometry=str(i[1]),
                )

            if len(react.products) > 0:
                listOfProducts = ET.SubElement(current_reaction, "listOfProducts")
            for i in react.products:
                ET.SubElement(
                    listOfProducts,
                    "speciesReference",
                    species=i[0],
                    stoichiometry=str(i[1]),
                )

            if len(react.modifiers) > 0:
                listOfModifiers = ET.SubElement(current_reaction, "listOfModifiers")
            for i in react.modifiers:
                ET.SubElement(listOfModifiers, "modifierSpeciesReference", species=i)

            current_reaction.append(react.kinetic_law)

        return sbml

    # #This can be changed later to no longer pretty print
    # def write_to_file(self,file_name,pretty_print = True):
    # 	# sbml = self.create_xml()
    # 	# et = ET.tostring(sbml,encoding='utf-8')
    # 	# dom = xml.dom.minidom.parseString(et)
    # 	# dom_pretty = dom.toprettyxml(encoding='utf-8')
    # 	with open(file_name,"w") as f:
    # 		if pretty_print:
    # 			f.write(self.dump(pretty_print = pretty_print))
    # 		else:
    # 			f.write('<?xml version="1.0" encoding="utf-8"?>')
    # 			f.write(ET.tostring(self.create_xml()).decode('utf-8'))

    # print out xml file
    def dump(self, pretty_print=True):
        sbml = self.create_xml()
        et = ET.tostring(sbml)
        dom = xml.dom.minidom.parseString(et)
        dom_pretty = dom.toprettyxml(encoding="utf-8").decode("utf-8")
        if pretty_print:
            return dom_pretty
        else:
            return '<?xml version="1.0" encoding="utf-8"?>' + et.decode("utf-8")


def sbml(segment, filename=None, model_name=None, pretty=True):
    rxd.initializer._do_init()
    section = segment.sec
    if model_name is not None:
        output = middle_man(model_name)
    else:
        output = middle_man(section.name())
    regions = section.psection()["regions"]
    species = section.psection()["species"]

    for i in regions:
        node = None
        for j in species:
            if i in j.regions or j.regions == [None]:
                node = j.nodes(segment)[0]
        output.add_compartment(i.name, size=node.volume)

    for i in species:
        if isinstance(i, rxd.Parameter):
            node = i.nodes(segment)[0]
            tempInitial = i.initial
            if i.initial is None:
                tempInitial = 0.0
            elif not isinstance(tempInitial, float) and not isinstance(
                tempInitial, int
            ):
                tempInitial = i.initial(node)
            output.add_parameter(i.name, i.initial)
            continue
        for j in i.regions:
            node = i.nodes(segment)[0]
            tempInitial = i.initial
            if i.initial is None:
                tempInitial = 0.0
            elif not isinstance(tempInitial, float) and not isinstance(
                tempInitial, int
            ):
                tempInitial = i.initial(node)
            output.add_species(
                name=i.name, compartment=j.name, initial_amount=tempInitial
            )

    for i in rxd.rxd._all_reactions:
        if isinstance(i, rxd.MultiCompartmentReaction):
            raise RxDException("Cannot export MultiCompartmentReactions")

    # MIGHT BE A PROBLEM WHEN MULTI COMPARTMENT REACTIONS
    reactions = [
        r()
        for r in rxd.rxd._all_reactions
        if (
            any(
                [
                    ((reg in r()._active_regions) or (reg is None))
                    for reg in r()._regions
                ]
            )
        )
        and isinstance(r(), rxd.Reaction)
    ]

    rates = [
        r()
        for r in rxd.rxd._all_reactions
        if (
            any(
                [
                    ((reg in r()._active_regions) or (reg is None))
                    for reg in r()._regions
                ]
            )
        )
        and isinstance(r(), rxd.Rate)
    ]

    num = 0
    for react in reactions:
        lhs = react._scheme._lhs._items
        rhs = react._scheme._rhs._items
        reacting_regions = []
        if react._regions == [None]:
            reacting_regions = regions
        else:
            reacting_regions = react._regions
        for reg in reacting_regions:
            reactants = [
                [key.name + "_" + reg.name, value]
                for key, value in zip(list(lhs.keys()), list(lhs.values()))
            ]
            products = [
                [key.name + "_" + reg.name, value]
                for key, value in zip(list(rhs.keys()), list(rhs.values()))
            ]

            kineticLaw = ET.Element("kineticLaw")
            math = ET.SubElement(
                kineticLaw, "math", xmlns="http://www.w3.org/1998/Math/MathML"
            )

            # adds the region to multiply it by
            region_multiple = ET.SubElement(math, "apply")
            ET.SubElement(region_multiple, "times")
            ET.SubElement(region_multiple, "ci").text = reg.name

            parameters_in_react = []

            recursive_search(
                react._rate_arithmeticed, region_multiple, parameters_in_react, reg.name
            )
            listOfParameters = ET.Element("listOfParameters")

            for i in parameters_in_react:
                if i[0] in output.parameters:
                    continue
                ET.SubElement(
                    listOfParameters, "parameter", id=i[0], name=i[1], value=str(i[2])
                )

            if list(listOfParameters):
                kineticLaw.append(listOfParameters)
            output.add_reaction(
                "Reaction_" + str(num),
                reversible=(react._dir == "<>"),
                products=products,
                reactants=reactants,
                kinetic_law=kineticLaw,
                modifiers=[],
            )
            num += 1

    num = 0
    for rate in rates:
        list_regions = []
        if rate._regions == [None]:
            list_regions = regions
        else:
            list_regions = rate._regions
        for r in list_regions:
            # math = ET.Element("apply")
            # ET.SubElement(math,"plus")
            # ET.SubElement(math,"cn").text = "0"
            # parameters = []
            # math_law = ex.recursive_search(rate._original_rate,math,parameters,r.name)
            # output.add_rate(rate._species().name + "_" + r.name,math)
            products = [[rate._species().name + "_" + r.name, 1]]
            reactants = []
            modifiers = [
                s().name + "_" + r.name
                for s in rate._involved_species
                if isinstance(s(), rxd.Parameter) == False
                and rate._species().name != s().name
            ]
            kineticLaw = ET.Element("kineticLaw")
            math = ET.SubElement(
                kineticLaw, "math", xmlns="http://www.w3.org/1998/Math/MathML"
            )
            region_multiple = ET.SubElement(math, "apply")
            ET.SubElement(region_multiple, "times")
            ET.SubElement(region_multiple, "ci").text = reg.name
            parameters = []
            recursive_search(rate._original_rate, region_multiple, parameters, r.name)
            output.add_reaction(
                "Rate_" + str(num),
                reversible=True,
                products=products,
                reactants=reactants,
                kinetic_law=kineticLaw,
                modifiers=modifiers,
            )
            num += 1

    output.add_unit_def("substance")
    output.add_unit("substance", "mole", 1, -3)
    output.add_unit_def("time")
    output.add_unit("time", "second", 1, -3)
    # output.add_unit_def("volume")
    # output.add_unit("volume","litre",1,1)

    final = output.dump(pretty_print=pretty)
    if filename:
        with open(filename, "w") as f:
            f.write(final)
    return final


function_names = {
    "acos": "arccos",
    "acosh": "arccosh",
    "asin": "arcsin",
    "asinh": "arcsinh",
    "atan": "arctanh",
    "atan2": "exception",  # FUNCTION 2
    "ceil": "ceiling",
    "copysign": "ERROR",
    "cos": "cos",
    "cosh": "cosh",
    "degrees": "ERROR",
    "erf": "ERROR",  # what do it mean
    "erfc": "ERROR",  # what do it mean
    "exp": "exp",
    "expm1": "exception",
    "fabs": "abs",
    "factorial": "ERROR",
    "floor": "floor",
    "fmod": "ERROR",
    "gamma": "eulergamma",
    "hypot": "exception",  # COULD IMPLEMENT FUNCTION 2
    "ldexp": "ERROR",
    "lgamma": "exception",  # what do it mean
    "log": "ln",
    "log10": "log",
    "log1p": "exception",
    "pow": "power",  # FUNCTION 2
    "radians": "ERROR",
    "sin": "sin",
    "sinh": "sinh",
    "sqrt": "exception",
    "tan": "tan",
    "tanh": "tanh",
    "trunc": "ERROR",
    "-": "exception",
}


def recursive_search(arth_obj, kineticLaw, parameters, compartment_name):
    items = []
    counts = []
    result = ""
    node = ET.SubElement(kineticLaw, "apply")

    for item, count in zip(
        list(arth_obj._items.keys()), list(arth_obj._items.values())
    ):
        if count:
            if isinstance(item, rxd.Parameter):
                parameters.append([item.name, item.name, item.initial])
                # ET.SubElement(parameters,"parameter", id=item.name,name=item.name,value=str(item.initial))
                temp = ET.Element("ci")
                temp.text = item.name
                items.append(temp)
                counts.append(count)
            elif isinstance(item, rxd.species._SpeciesMathable):
                temp = ET.Element("ci")
                if compartment_name == "":
                    temp.text = item.name
                else:
                    temp.text = item.name + "_" + compartment_name
                items.append(temp)
                counts.append(count)
            elif isinstance(item, int):
                temp = ET.Element("cn")
                temp.text = str(item)
                items.append(temp)
                counts.append(count)
            else:
                recursive_node = determine_type(item, parameters, compartment_name)
                items.append(recursive_node)
                counts.append(count)

    ET.SubElement(node, "plus")

    for i, c in zip(items, counts):
        tempNode = ET.SubElement(node, "apply")
        ET.SubElement(tempNode, "times")

        tempNode.append(i)
        ET.SubElement(tempNode, "cn").text = str(c)

    if len(items) == 1:
        ET.SubElement(node, "cn").text = "0"
    return node


def determine_type(obj, parameters, comp_name):
    tree = ET.Element("apply")
    if isinstance(obj, rxd.rxdmath._Product):
        ET.SubElement(tree, "times")
        recursive_search(obj._a, tree, parameters, comp_name)  # arithmeticed
        recursive_search(obj._b, tree, parameters, comp_name)  # arithmeticed
    elif isinstance(obj, rxd.rxdmath._Quotient):
        ET.SubElement(tree, "divide")
        recursive_search(obj._a, tree, parameters, comp_name)  # arithmeticed
        recursive_search(obj._b, tree, parameters, comp_name)  # arithmeticed
    elif isinstance(obj, rxd.rxdmath._Arithmeticed):
        recursive_search(obj, tree, parameters, comp_name)
    elif isinstance(obj, rxd.rxdmath._Function):
        if function_names[obj._fname] == "exception":
            # Do Something bro
            if obj._fname == "-":
                ET.SubElement(tree, "times")
                recursive_search(obj._obj, tree, parameters, comp_name)
                ET.SubElement(tree, "cn").text = str(-1)
        elif function_names[obj._fname] == "ERROR":
            RxDException(obj._fname + " is not supported in this context")
        else:
            ET.SubElement(tree, function_names[obj._fname])
            recursive_search(obj._obj, tree, parameters, comp_name)
    elif isinstance(obj, rxd.rxdmath._Function2):
        if function_names[obj._fname] == "exception":
            if obj._fname == "atan2":
                temp = ET.SubElement(tree, function_names["atan"])
                applytemp = ET.SubElement(temp, "apply")
                temp2 = ET.SubElement(temp, "divide")
                recursive_search(obj._obj1, applytemp, parameters, comp_name)
                recursive_search(obj._obj2, applytemp, parameters, comp_name)
        elif function_names[obj._fname] == "ERROR":
            RxDException(obj._fname + " is not supported in this context")
        else:
            ET.SubElement(tree, function_names[obj._fname])
            recursive_search(obj._obj1, tree, parameters, comp_name)
            recursive_search(obj._obj2, tree, parameters, comp_name)

    return tree



import sys
import ctypes
import neuron
import numpy
from neuron import h
import traceback

# reducing the indirection improves performance
try:
    _numpy_core_multiarray_int_asbuffer = numpy.core.multiarray.int_asbuffer
except:
    pass
_ctypes_addressof = ctypes.addressof
_numpy_array = numpy.array
_numpy_frombuffer = numpy.frombuffer

#
# connect to dll via ctypes
#

nrn_dll_sym = neuron.nrn_dll_sym

# declare prototype
nonvint_block_prototype = ctypes.CFUNCTYPE(
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.POINTER(ctypes.c_double),
    ctypes.POINTER(ctypes.c_double),
    ctypes.c_int,
)

set_nonvint_block = nrn_dll_sym("set_nonvint_block")
set_nonvint_block.argtypes = [nonvint_block_prototype]
unset_nonvint_block = nrn_dll_sym("unset_nonvint_block")
unset_nonvint_block.argtypes = [nonvint_block_prototype]

# Some info not available from the HocObject
v_structure_change = nrn_dll_sym("v_structure_change", ctypes.c_int)

# items in call are each a list of 10 callables
#    [setup, initialize, # method 0, 1
#    current, conductance, fixed_step_solve, # method 2, 4
#    ode_count, ode_reinit, ode_fun, ode_solve, ode_jacobian, ode_abs_tolerance ] # method 5-10
call = []

# e.g.
test = """
def setup(): #0
  print("setup")
def initialize(): #1
  print("initialize")
def current(rhs): #2
  print("outward current to be subtracted from rhs")
def conductance(d): #3
  print("conductance to be added to d")
def fixed_step_solve(dt): #4
  print("fixed_step_solve")
def ode_count(offset): #5
  print("ode_count") ; return 0
def ode_reinit(y): #6 # fill y with initial values of states
  print("ode_reinit")
def ode_fun(t, y, ydot): #7 # from t, y, determine ydot
  print("ode_fun")
def ode_solve(dt, t, b, y): #8 #solve mx=b replace b with x (y available if m depends on it
  print("ode_solve")
def ode_jacobian(dt, t, ypred, fpred): #9 optionally prepare jacobaian for fast ode_solve.
  print("ode_jacobian %g %g" % (t, dt))
def ode_abs_tolerance(y_abs_tolerance): #10 fill with cvode.atol()*scalefactor
  print("ode_abs_tolerance") # on entry, y_abs_tolerance filled with cvode.atol()

call.append([
    setup, initialize, # method 0, 1
    current, conductance, fixed_step_solve, # method 2, 3, 4
    ode_count, ode_reinit, ode_fun, ode_solve, ode_jacobian, # method 5-9
    ode_abs_tolerance
  ])
"""

ode_count_method_index = 5


def register(c):
    unregister(c)
    call.append(c)
    activate_callback(True)


def unregister(c):
    v_structure_change.value = 1
    if c in call:
        call.remove(c)
    if len(call) == 0:
        activate_callback(False)


def clear():
    while len(call):
        unregister(call[0])


def ode_count_all(offset):
    global nonvint_block_offset
    nonvint_block_offset = offset
    cnt = 0
    for c in call:
        if c[ode_count_method_index]:
            cnt += c[ode_count_method_index](nonvint_block_offset + cnt)
    # print("ode_count_all %d, offset=%d\n"%(cnt, nonvint_block_offset))
    return cnt


pc = h.ParallelContext()
_pc_dt = pc.dt
_pc_t = pc.t

# see nrn/src/nrnoc/nonvintblock.h for the magic method numbers

_empty_array = _numpy_array([])

_no_args = (0, 1)
_pd1_arg = (2, 3, 6, 10)
_float_size = numpy.dtype(float).itemsize


def numpy_from_pointer(cpointer, size):
    buf_from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    buf_from_mem.restype = ctypes.py_object
    buf_from_mem.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
    cbuffer = buf_from_mem(cpointer, size * _float_size, 0x200)
    return numpy.ndarray((size,), float, cbuffer, order="C")


def nonvint_block(method, size, pd1, pd2, tid):
    # print('nonvint_block called with method = %d l=%d tid=%d' % (method,size,tid))
    try:
        assert tid == 0
        rval = 0
        if method == ode_count_method_index:
            rval = ode_count_all(
                size
            )  # count of the extra states-equations managed by us
        else:
            if pd1:
                if size:
                    pd1_array = numpy_from_pointer(pd1, size)
                else:
                    pd1_array = _empty_array
            else:
                pd1_array = None
            if pd2:
                if size:
                    pd2_array = numpy_from_pointer(pd2, size)
                else:
                    pd2_array = _empty_array
            else:
                pd2_array = None

            if method in _no_args:
                args = ()
            elif method in _pd1_arg:
                args = (pd1_array,)
            elif method == 4:
                args = (_pc_dt(tid),)
            elif method == 7:
                args = (_pc_t(tid), pd1_array, pd2_array)
            elif method in (8, 9):
                args = (_pc_dt(tid), _pc_t(tid), pd1_array, pd2_array)
            for c in call:
                if c[method] is not None:
                    c[method](*args)
    except:
        traceback.print_exc()
        rval = -1
    return rval


_callback = nonvint_block_prototype(nonvint_block)


def activate_callback(activate):
    if activate:
        set_nonvint_block(_callback)
    else:
        unset_nonvint_block(_callback)


if __name__ == "__main__":
    exec(test)  # see above string

    s = h.Section()
    print("fixed step finitialize")
    h.finitialize(0)
    print("fixed step fadvance")
    h.fadvance()

    h.load_file("stdgui.hoc")
    print("cvode active")
    h.cvode_active(1)
    print("cvode step finitialize")
    h.finitialize(0)
    print("cvode fadvance")
    h.fadvance()



"""test of pure diffusion

Robert A McDougal
March 2013 - June 2014
"""

from neuron import h, crxd as rxd
import numpy
import __main__

name = __main__.__file__
if name[-3:] == ".py":
    name = name[:-3]

h.load_file("stdrun.hoc")
h.cvode_active(True)
dend1 = h.Section()
dend1.diam = 2
dend1.nseg = 101
dend1.L = 50

dend2 = h.Section()
dend2.diam = 2
dend2.nseg = 101
dend2.L = 50
dend2.connect(dend1)

diff_constant = 1

r = rxd.Region(h.allsec(), dx=0.25)

rxd.set_solve_type([dend1], dimension=3)

ca = rxd.Species(
    r,
    d=diff_constant,
    initial=lambda node: 1
    if (0.8 < node.x and node.segment in dend1)
    or (node.x < 0.2 and node.segment in dend2)
    else 0,
)
h.finitialize()
initial_amount = (
    numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)
).sum()

if __name__ == "__main__":
    from matplotlib import pyplot

    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)
        pyplot.plot(
            [nd.x for nd in ca.nodes if nd.segment in dend1]
            + [1 + nd.x for nd in ca.nodes if nd.segment in dend2],
            [nd.concentration for nd in ca.nodes if nd.segment in dend1]
            + [nd.concentration for nd in ca.nodes if nd.segment in dend2],
            ".",
        )
    pyplot.tight_layout()
    pyplot.savefig("{0}.png".format(name))
    pyplot.show()
else:
    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)

final_amount = (
    numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)
).sum()

print(
    "initial {}\tfinal {}\tdiff {}".format(
        initial_amount, final_amount, initial_amount - final_amount
    )
)



import distutils.util
import os

from neuron import h, gui


def test_direct_memory_transfer():
    h("""create soma""")
    h.soma.L = 5.6419
    h.soma.diam = 5.6419
    h.soma.insert("hh")
    ic = h.IClamp(h.soma(0.5))
    ic.delay = 0.5
    ic.dur = 0.1
    ic.amp = 0.3

    # for testing external mod file
    h.soma.insert("Sample")

    h.cvode.use_fast_imem(1)
    h.cvode.cache_efficient(1)

    v = h.Vector()
    v.record(h.soma(0.5)._ref_v, sec=h.soma)
    i_mem = h.Vector()
    i_mem.record(h.soma(0.5)._ref_i_membrane_, sec=h.soma)
    tv = h.Vector()
    tv.record(h._ref_t, sec=h.soma)
    h.run()
    vstd = v.cl()
    tvstd = tv.cl()
    i_memstd = i_mem.cl()
    # Save current (after run) value to compare with transfer back from coreneuron
    tran_std = [h.t, h.soma(0.5).v, h.soma(0.5).hh.m]

    from neuron import coreneuron

    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.model_stats = True
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )
    coreneuron.num_gpus = 1

    pc = h.ParallelContext()

    def run(mode):
        pc.set_maxstep(10)
        h.stdinit()
        if mode == 0:
            pc.psolve(h.tstop)
        elif mode == 1:
            while abs(h.t - h.tstop) > 0.1 * h.dt:
                pc.psolve(h.t + 1.0)
        else:
            while abs(h.t - h.tstop) > 0.1 * h.dt:
                h.continuerun(h.t + 0.5)
                pc.psolve(h.t + 0.5)
        tran = [h.t, h.soma(0.5).v, h.soma(0.5).hh.m]

        assert tv.eq(tvstd)
        assert (
            v.cl().sub(vstd).abs().max() < 1e-10
        )  # usually v == vstd, some compilers might give slightly different results
        assert i_mem.cl().sub(i_memstd).abs().max() < 1e-10
        assert h.Vector(tran_std).sub(h.Vector(tran)).abs().max() < 1e-10

    for mode in [0, 1, 2]:
        run(mode)

    cnargs = coreneuron.nrncore_arg(h.tstop)
    h.stdinit()
    assert tv.size() == 1 and tvstd.size() != 1
    pc.nrncore_run(cnargs, 1)
    assert tv.eq(tvstd)

    # print warning if HocEvent on event queue when CoreNEURON starts
    def test_hoc_event():
        print("in test_hoc_event() at t=%g" % h.t)
        if h.t < 1.001:
            h.CVode().event(h.t + 1.0, test_hoc_event)

    fi = h.FInitializeHandler(2, test_hoc_event)
    coreneuron.enable = False
    run(0)
    coreneuron.enable = True
    run(0)


if __name__ == "__main__":
    test_direct_memory_transfer()
    h.quit()



from neuron import h
from neuron.units import ms, mV

h("objref po")
h.po = {}

import numpy as np
import subprocess

pc = h.ParallelContext()


# start fresh with respect to SaveState and BBSaveState
def rmfiles():
    if pc.id() == 0:
        subprocess.run("rm -f state*.bin", shell=True)
        subprocess.run("rm -r -f bbss_out", shell=True)
        subprocess.run("rm -r -f in", shell=True)
    pc.barrier()


rmfiles()


class Cell:
    def __init__(self, gid, x, y, z, theta):
        self._gid = gid
        self._setup_morphology()
        self.all = self.soma.wholetree()
        self._setup_biophysics()
        self.x = self.y = self.z = 0
        h.define_shape()
        self._rotate_z(theta)
        self._set_position(x, y, z)

        self._spike_detector = h.NetCon(self.soma(0.5)._ref_v, None, sec=self.soma)
        self.spike_times = h.Vector()
        self._spike_detector.record(self.spike_times)

        self._ncs = []

        self.soma_v = h.Vector().record(self.soma(0.5)._ref_v)

    def __repr__(self):
        return "{}[{}]".format(self.name, self._gid)

    def _set_position(self, x, y, z):
        for sec in self.all:
            for i in range(sec.n3d()):
                sec.pt3dchange(
                    i,
                    x - self.x + sec.x3d(i),
                    y - self.y + sec.y3d(i),
                    z - self.z + sec.z3d(i),
                    sec.diam3d(i),
                )
        self.x, self.y, self.z = x, y, z

    def _rotate_z(self, theta):
        """Rotate the cell about the Z axis."""
        for sec in self.all:
            for i in range(sec.n3d()):
                x = sec.x3d(i)
                y = sec.y3d(i)
                c = h.cos(theta)
                s = h.sin(theta)
                xprime = x * c - y * s
                yprime = x * s + y * c
                sec.pt3dchange(i, xprime, yprime, sec.z3d(i), sec.diam3d(i))


class BallAndStick(Cell):
    name = "BallAndStick"

    def _setup_morphology(self):
        self.soma = h.Section(name="soma", cell=self)
        self.dend = h.Section(name="dend", cell=self)
        self.dend.connect(self.soma)
        self.soma.L = self.soma.diam = 12.6157
        self.dend.L = 200
        self.dend.diam = 1

    def _setup_biophysics(self):
        for sec in self.all:
            sec.Ra = 100  # Axial resistance in Ohm * cm
            sec.cm = 1  # Membrane capacitance in micro Farads / cm^2
        self.soma.insert("hh")
        for seg in self.soma:
            seg.hh.gnabar = 0.12  # Sodium conductance in S/cm2
            seg.hh.gkbar = 0.036  # Potassium conductance in S/cm2
            seg.hh.gl = 0.0003  # Leak conductance in S/cm2
            seg.hh.el = -54.3  # Reversal potential in mV
        # Insert passive current in the dendrite
        self.dend.insert("pas")
        for seg in self.dend:
            seg.pas.g = 0.001  # Passive conductance in S/cm2
            seg.pas.e = -65  # Leak reversal potential mV

        self.syn = h.ExpSyn(self.dend(0.5))
        self.syn.tau = 2 * ms


class Ring:
    """A network of *N* ball-and-stick cells where cell n makes an
    excitatory synapse onto cell n + 1 and the last, Nth cell in the
    network projects to the first cell.
    """

    def __init__(
        self,
        N=5,
        stim_w=0.04,
        stim_t=9,
        stim_delay=1,
        syn_w=0.01,
        syn_delay=25,
        r=50,
    ):
        """
        :param N: Number of cells.
        :param stim_w: Weight of the stimulus
        :param stim_t: time of the stimulus (in ms)
        :param stim_delay: delay of the stimulus (in ms)
        :param syn_w: Synaptic weight
        :param syn_delay: Delay of the synapse
        :param r: radius of the network
        """
        self._N = N
        self.set_gids()  ### assign gids to processors
        self._syn_w = syn_w
        self._syn_delay = syn_delay
        self._create_cells(r)
        self._connect_cells()
        ### stimulate gid 0
        if pc.gid_exists(0):
            self._netstim = h.NetStim()
            self._netstim.number = 1
            self._netstim.start = stim_t
            self._nc = h.NetCon(
                self._netstim, pc.gid2cell(pc.id()).syn
            )  ### grab cell with gid==0 wherever it exists
            self._nc.delay = stim_delay
            self._nc.weight[0] = stim_w

    def set_gids(self):
        """Set the gidlist on this host."""
        #### Round-robin counting.
        #### Each host has an id from 0 to pc.nhost() - 1.
        self.gidlist = list(range(pc.id(), self._N, pc.nhost()))
        for gid in self.gidlist:
            pc.set_gid2node(gid, pc.id())

    def _create_cells(self, r):
        self.cells = {}
        for i in self.gidlist:  ### only create the cells that exist on this host
            theta = i * 2 * h.PI / self._N
            self.cells[i] = BallAndStick(
                i, h.cos(theta) * r, h.sin(theta) * r, 0, theta
            )

        ### associate the cell with this host and gid
        for cell in self.cells.values():
            pc.cell(cell._gid, cell._spike_detector)

    def _connect_cells(self):
        for target in self.cells.values():
            source_gid = (target._gid - 1 + self._N) % self._N
            nc = pc.gid_connect(source_gid, target.syn)
            nc.weight[0] = self._syn_w
            nc.delay = self._syn_delay
            target._ncs.append(nc)


class StarNet:
    """NetStim -> Cell -> N Cells -> Cell"""

    def __init__(self, n):
        self.ncell = [1, n, 1]
        self.nlayer = len(self.ncell)
        self.cells = {}
        # make cells
        for ilayer in range(self.nlayer):
            for icell in range(self.ncell[ilayer]):
                gid = self.info2gid(ilayer, icell)
                if (gid % pc.nhost()) == pc.id():
                    cell = BallAndStick(gid, float(icell), float(ilayer), 0.0, 0.0)
                    self.cells[gid] = cell
                    cell.ilayer = ilayer
                    cell.icell = icell
                    pc.set_gid2node(gid, pc.id())
                    pc.cell(gid, self.cells[gid]._spike_detector)
        # make connections (all to all from layer i-1 to layer i)
        self.nclist = {}
        for gid, cell in self.cells.items():
            if cell.ilayer > 0:
                src_ilayer = cell.ilayer - 1
                for src_icell in range(self.ncell[src_ilayer]):
                    srcgid = self.info2gid(src_ilayer, src_icell)
                    nc = pc.gid_connect(srcgid, cell.syn)
                    nc.weight[0] = 0.01
                    nc.delay = 20
                    self.nclist[(srcgid, gid)] = nc
        # stimulate gid 0 with NetStim
        if 0 in self.cells:
            self.ns = h.NetStim()
            self.ncstim = h.NetCon(self.ns, self.cells[0].syn)
            ns = self.ns
            ns.start = 6
            ns.interval = 10
            ns.number = 100
            nc = self.ncstim
            nc.delay = 2
            nc.weight[0] = 0.01
        # For some extra coverage of BBSaveState::node01
        if 100 in self.cells:
            cell = self.cells[100]
            self.xsyns = [h.ExpSyn(seg) for seg in cell.dend.allseg()]
            self.xnc = []
            for syn in self.xsyns:
                nc = pc.gid_connect(0, syn)
                nc.delay = 20
                nc.weight[0] = 0.0001
                self.xnc.append(nc)

    def info2gid(self, ilayer, icell):
        return ilayer * 100 + icell

    def gid2info(self, gid):
        return (int(gid / 100)), gid % 100

    def topol(self):
        print(pc.id(), self.cells)
        print(pc.id(), self.nclist)


out2in_sh = r"""
#!/usr/bin/env bash
out=bbss_out
rm -f in/*
mkdir -p in
cat $out/tmp > in/tmp
for f in $out/tmp.*.* ; do
  i=`echo "$f" | sed 's/.*tmp\.\([0-9]*\)\..*/\1/'`
  if test ! -f in/tmp.$i ; then
    cnt=`ls $out/tmp.$i.* | wc -l`
    echo $cnt > in/tmp.$i
    cat $out/tmp.$i.* >> in/tmp.$i
  fi
done
"""


def cp_out_to_in():
    if pc.id() == 0:
        import tempfile

        with tempfile.NamedTemporaryFile("w") as scriptfile:
            scriptfile.write(out2in_sh)
            scriptfile.flush()
            subprocess.check_call(["/bin/bash", scriptfile.name])

    pc.barrier()


def prun(tstop, restore=False):
    pc.set_maxstep(10 * ms)
    h.finitialize(-65 * mV)

    if restore == "SaveState":
        ns = h.SaveState()
        sf = h.File("state%d.bin" % pc.id())
        ns.fread(sf)
        ns.restore(0)  # event queue restored
        sf.close()
    elif restore == "BBSaveState":
        cp_out_to_in()  # prepare for restore.
        bbss = h.BBSaveState()
        bbss.restore_test()
    else:
        pc.psolve(tstop / 2)

        # SaveState save
        ss = h.SaveState()
        ss.save()
        sf = h.File("state%d.bin" % pc.id())
        ss.fwrite(sf)
        sf.close()

        # BBSaveState Save
        cnt = h.List("PythonObject").count()
        for i in range(1):
            bbss = h.BBSaveState()
            bbss.save_test()
            bbss = None
        assert h.List("PythonObject").count() == cnt

    pc.psolve(tstop)


def get_all_spikes(ring):
    local_data = {cell._gid: list(cell.spike_times) for cell in ring.cells.values()}
    all_data = pc.py_allgather([local_data])

    pc.barrier()
    pc.done()

    data = {}
    for d in all_data:
        data.update(d[0])

    return data


def compare_dicts(dict1, dict2):
    # assume dict is {gid:[spiketimes]}

    # In case iteration order not same in dict1 and dict2, use dict1 key
    # order to access dict
    keylist = dict1.keys()

    # verify same set of keys
    assert set(keylist) == set(dict2.keys())

    # verify same count of spikes for each key
    assert [len(dict1[k]) for k in keylist] == [len(dict2[k]) for k in keylist]

    # Put spike times in array so can compare with a tolerance.
    array_1 = np.array([val for k in keylist for val in dict1[k]])
    array_2 = np.array([val for k in keylist for val in dict2[k]])
    if not np.allclose(array_1, array_2):
        print(array_1)
        print(array_2)
        print(array_1 - array_2)
    assert np.allclose(array_1, array_2)


def test_bas():

    # h.execute1(...) does not call mpi_abort on failure
    assert h.execute1("1/0") == 0
    assert h.execute1("2/0", 0) == 0  # no error message printed

    # MPI_Abort can be avoided on hoc errors.
    oldflag = pc.mpiabort_on_error(0)
    assert h("""3/0""") == 0
    try:
        x = h.log(-1)
        assert False
    except:
        assert True
    pc.mpiabort_on_error(oldflag)

    stdspikes = {
        0: [10.925000000099914, 143.3000000001066],
        1: [37.40000000009994, 169.7750000000825],
        2: [63.87500000010596, 196.25000000005844],
        3: [90.35000000011198],
        4: [116.825000000118],
    }

    stdspikes_after_100 = {}
    for gid in stdspikes:
        stdspikes_after_100[gid] = [spk_t for spk_t in stdspikes[gid] if spk_t >= 100.0]

    ring = Ring()

    prun(200 * ms)  # at tstop/2 does a SaveState.save and BBSaveState.save
    stdspikes = get_all_spikes(ring)
    stdspikes_after_100 = {}
    for gid in stdspikes:
        stdspikes_after_100[gid] = [spk_t for spk_t in stdspikes[gid] if spk_t >= 100.0]
    compare_dicts(get_all_spikes(ring), stdspikes)

    prun(200 * ms, "SaveState")  # SaveState restore to start at t = tstop/2
    compare_dicts(get_all_spikes(ring), stdspikes_after_100)

    prun(200 * ms, "BBSaveState")  # BBSaveState restore to start at t = tstop/2
    compare_dicts(get_all_spikes(ring), stdspikes_after_100)


def test_starnet():
    pc.gid_clear()
    starnet = StarNet(8)
    starnet.topol()
    tstop = 100.0
    prun(tstop)
    stdspikes = get_all_spikes(starnet)
    stdspikes_half = {}
    for gid in stdspikes:
        stdspikes_half[gid] = [spk_t for spk_t in stdspikes[gid] if spk_t >= tstop / 2]
    prun(tstop, "BBSaveState")  # BBSaveState restore to start at t = tstop/2
    compare_dicts(get_all_spikes(starnet), stdspikes_half)

    # test for binq mode.
    h.CVode().queue_mode(1)
    prun(tstop)
    compare_dicts(get_all_spikes(starnet), stdspikes)
    prun(tstop, "BBSaveState")
    compare_dicts(get_all_spikes(starnet), stdspikes_half)

    h.dt = 1.0 / 64.0  # issue 1480
    prun(tstop)
    stdspikes = get_all_spikes(starnet)
    stdspikes_half = {}
    for gid in stdspikes:
        stdspikes_half[gid] = [spk_t for spk_t in stdspikes[gid] if spk_t >= tstop / 2]

    prun(tstop, "BBSaveState")
    compare_dicts(get_all_spikes(starnet), stdspikes_half)
    h.CVode().queue_mode(0)
    h.dt = 0.025


if __name__ == "__main__":
    test_bas()
    test_starnet()
    pc.barrier()
    h.quit()



import numpy
import bisect


def _segment_3d_pts(sec):
    """
    given a section, return a list of (xs, ys, zs, diams) for each segment

    .. warning::

        Does not currently support non-frustum based sections (i.e. no support
        for new 3d styles, like soma outlines)

    .. warning::

        This assumes a 3d style exists. The safest way to call this is to call
        h.define_shape() first
    """
    # TODO: fix the issue described in the warning
    #       (when this was written, these objects were only under development)

    # this is based on neuron.rxd.dimension3.centroids_by_segment(sec)

    n3d = sec.n3d()
    length = sec.L

    arc3d = [sec.arc3d(i) for i in range(n3d)]
    x3d = numpy.array([sec.x3d(i) for i in range(n3d)])
    y3d = numpy.array([sec.y3d(i) for i in range(n3d)])
    z3d = numpy.array([sec.z3d(i) for i in range(n3d)])
    diam3d = numpy.array([sec.diam3d(i) for i in range(n3d)])

    dx = length / sec.nseg
    result = []
    for i in range(sec.nseg):
        x_lo = i * dx
        x_hi = (i + 1) * dx
        pts = [x_lo] + _values_strictly_between(x_lo, x_hi, arc3d) + [x_hi]
        local_x3d = list(numpy.interp(pts, arc3d, x3d))
        local_y3d = list(numpy.interp(pts, arc3d, y3d))
        local_z3d = list(numpy.interp(pts, arc3d, z3d))
        local_diam3d = list(numpy.interp(pts, arc3d, diam3d))
        result.append((local_x3d, local_y3d, local_z3d, local_diam3d, pts))
    return result


def _values_between(lo, hi, data):
    i_lo = bisect.bisect_left(data, lo)
    i_hi = bisect.bisect_right(data, hi)
    return data[i_lo:i_hi]


def _values_strictly_between(lo, hi, data):
    temp = _values_between(lo, hi, data)
    if temp and temp[0] == lo:
        temp = temp[1:]
    if temp and temp[-1] == hi:
        temp = temp[:-1]
    return temp



from .surface import surface
from .triangularMesh import TriangularMesh
from .voxelize import voxelize

# from .voxelize2 import voxelize2
from .scalarField import ScalarField
from .FullJoinMorph import fullmorph as voxelize2



import inspect
import itertools
import os

import numpy


tol = float(os.environ.get("NRN_RXD_TEST_TOLERANCE", "1e-10"))
dt_eps = 1e-20


def get_data_file_name(frame):
    """returns the filename for the data file need for the test."""

    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 4)
    testfunc_name = calframe[frame][3]
    assert testfunc_name.startswith("test_")
    return testfunc_name[5:] + ".dat"


def get_correct_data_for_test():
    """returns a path to the file with the correct data for a test."""

    data_filename = get_data_file_name(frame=3)
    basepath = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(basepath, "testdata", "test", data_filename)


def save_data_from_test(save_path):
    """save the data generated by a test."""
    basepath = os.path.abspath(save_path)
    if not os.path.exists(basepath):
        os.mkdir(basepath)
    filepath = os.path.join(basepath, get_data_file_name(frame=4))
    return filepath


def collect_data(h, rxd, data, save_path, num_record=10):
    """grabs the membrane potential data, h.t, and the rxd state values"""

    data["record_count"] += 1
    if data["record_count"] > num_record:
        h.stoprun = True
        return
    all_potentials = [seg.v for seg in itertools.chain.from_iterable(h.allsec())]
    rxd_1d = list(rxd.node._states)
    rxd_3d = []
    rxd_ecs = []
    for sp in rxd.species._all_species:
        s = sp()
        if s and hasattr(s, "_intracellular_instances"):
            for ics in s._intracellular_instances.values():
                rxd_3d += list(ics.states)
        if s and hasattr(s, "_extracellular_instances"):
            for ecs in s._extracellular_instances.values():
                rxd_ecs += list(ecs.states.flatten())
    all_rxd = rxd_1d + rxd_3d + rxd_ecs
    local_data = [h.t] + all_potentials + all_rxd

    # remove data before t=0
    if h.t == 0:
        data["data"] = []
        data["record_count"] = 1
    # remove previous record if h.t is the same
    if data["record_count"] > 1 and h.t == data["data"][-len(local_data)]:
        data["record_count"] -= 1
        del data["data"][-len(local_data) :]
    # add new data record
    data["data"].extend(local_data)
    if data["record_count"] == 2:
        data["rlen"] = len(local_data)

    # save the test data (if the --save option was used)
    if save_path:
        with open(save_data_from_test(save_path), "wb") as f:
            numpy.array(data["data"]).tofile(f)


def compare_data(data):
    """compares the test data with the correct data"""

    rlen = data["rlen"]
    corr_dat = numpy.fromfile(get_correct_data_for_test()).reshape(-1, rlen)
    tst_dat = numpy.array(data["data"]).reshape(-1, rlen)
    t1 = corr_dat[:, 0]
    t2 = tst_dat[:, 0]
    # remove any initial t that are greter than the next t
    # (removes times before 0) in correct data
    c = 0
    while c < len(t1) - 1 and t1[c] > t1[c + 1]:
        c = c + 1
    t1 = numpy.delete(t1, range(c))
    corr_dat = numpy.delete(corr_dat, range(c), 0)
    # remove any initial t that are greter than the next t
    # (removes times before 0) in test data
    c = 0
    while c < len(t2) - 1 and t2[c] > t2[c + 1]:
        c = c + 1
    t2 = numpy.delete(t2, range(c))
    tst_dat = numpy.delete(tst_dat, range(c), 0)
    # get rid of repeating t in correct data (otherwise interpolation fails)
    c = 0
    while c < len(t1) - 1:
        c1 = c + 1
        while c1 < len(t1) and abs(t1[c] - t1[c1]) < dt_eps:
            c1 = c1 + 1
        t1 = numpy.delete(t1, range(c, c1 - 1))
        corr_dat = numpy.delete(corr_dat, range(c, c1 - 1), 0)
        c = c + 1
    # get rid of the test data outside of the correct data time interval
    t2_n = len(t2)
    t2_0 = 0
    while t2[t2_n - 1] > t1[-1]:
        t2_n = t2_n - 1
    while t2[t2_0] < t1[0]:
        t2_0 = t2_0 + 1
    # interpolate and compare
    corr_vals = numpy.array(
        [numpy.interp(t2[t2_0:t2_n], t1, corr_dat[:, i].T) for i in range(1, rlen)]
    )
    max_err = numpy.amax(abs(corr_vals.T - tst_dat[t2_0:t2_n, 1:]))
    return max_err



# Test proper management of HOC PythonObject (no memory leaks)
# Expanded for more testing and coverage of BBSaveState.

from neuron import h
from neuron.units import ms, mV

import numpy as np
import subprocess


def hlist():
    return h.List("PythonObject")


def test_1():
    hl = hlist()
    npo = hl.count()  # compare against this. Previous tests may leave some.

    h("objref po")
    h.po = {}
    assert hl.count() == (npo + 1)
    h.po = h.Vector()
    assert hl.count() == npo
    h.po = None
    assert hl.count() == npo

    h("po = new PythonObject()")
    assert hl.count() == (npo + 1)
    h("po = po.list()")
    assert type(h.po) == type([])
    assert hl.count() == (npo + 1)
    h("objref po")
    assert hl.count() == npo

    h(
        """obfunc test_hoc_po() { localobj foo
         foo = new PythonObject()
         return foo.list()
       }
       po = test_hoc_po()
    """
    )
    assert hl.count() == (npo + 1)
    h.po = None
    assert hl.count() == npo
    assert type(h.test_hoc_po()) == type([])
    assert hl.count() == npo

    # nrn.Section tests
    for sec in h.allsec():
        h.delete_section(sec=sec)

    h("""create soma""")
    h.po = h.soma
    assert h.po == h.soma
    assert hl.count() == (npo + 1)
    h.po = None
    assert hl.count() == npo
    h.po = h.soma
    h.delete_section(sec=h.soma)
    assert str(h.po) == "<deleted section>"
    h.po = None
    assert hl.count() == npo

    h.po = h.Section(name="pysoma")
    h("po.L = 5")
    assert h.po.L == 5
    assert hl.count() == (npo + 1)
    h.po = None
    assert hl.count() == npo
    assert len([sec for sec in h.allsec()]) == 0


pc = h.ParallelContext()


class PyCell:
    def __init__(self, gid):
        s = {i: h.Section(name=i, cell=self) for i in ["soma", "dend", "axon"]}
        s["dend"].connect(s["soma"](0))
        s["axon"].connect(s["soma"](1))
        self.soma = s["soma"]
        a = s["axon"]
        self.sections = s
        for s in self.sections.values():
            s.L = 10
            s.diam = 10 if "soma" in s.name() else 1
            s.insert("hh")

        self.syn = h.ExpSyn(self.sections["dend"](0.5))
        self.syn.e = 0
        self.gid = gid
        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(a(1)._ref_v, None, sec=a))


h(
    """
objref pc
pc = new ParallelContext()
begintemplate TestHocPOCell
  public soma, dend, axon, gid, syn
  external pc
  create soma, dend, axon
  objref syn
  proc init() { localobj nc, nil
    connect dend(0), soma(0)
    connect axon(0), soma(1)
    forall { insert hh  L=10  diam = 1 }
    soma.diam = 10

    dend {syn = new ExpSyn(.5)}
    syn.e = 0
    gid = $1
    pc.set_gid2node(gid, pc.id())
    axon nc = new NetCon(&v(1), nil)
    pc.cell(gid, nc)
  }
endtemplate TestHocPOCell
"""
)


class Ring:
    def __init__(self, npy, nhoc):
        pc.gid_clear()
        self.ngid = npy + nhoc
        self.cells = {gid: PyCell(gid) for gid in range(npy)}
        self.cells.update({gid: h.TestHocPOCell(gid) for gid in range(npy, self.ngid)})
        self.mkcon(self.ngid)
        self.mkstim()
        self.spiketime = h.Vector()
        self.spikegid = h.Vector()
        pc.spike_record(-1, self.spiketime, self.spikegid)
        for gid in range(self.ngid):
            pc.threshold(gid, -10.0)

    def mkcon(self, ngid):
        self.ncs = [
            pc.gid_connect(gid, self.cells[(gid + 1) % ngid].syn) for gid in range(ngid)
        ]
        for nc in self.ncs:
            nc.delay = 1.0
            nc.weight[0] = 0.004

    def mkstim(self):
        self.ic = h.IClamp(self.cells[0].soma(0.5))
        self.ic.delay = 1
        self.ic.dur = 0.2
        self.ic.amp = 0.3


def test_2():
    hl = hlist()
    npo = hl.count()

    r = Ring(3, 2)
    assert hl.count() == npo
    for gid in r.cells:
        assert pc.gid2cell(gid) == r.cells[gid]
    assert hl.count() == npo
    pc.gid_clear()


# BBSaveState for mixed (hoc and python cells) Ring.

# some helpers copied from ../parallel_tests/test_bas.py
def subprocess_run(cmd):
    subprocess.run(cmd, shell=True).check_returncode()


def rmfiles():
    if pc.id() == 0:
        subprocess_run("rm -r -f bbss_out")
        subprocess_run("rm -r -f in")
        subprocess_run("rm -r -f binbufout")
        subprocess_run("rm -r -f binbufin")
        subprocess_run("mkdir binbufout")
        subprocess_run("rm -f allcell-bbss.dat")
    pc.barrier()


def cp_out_to_in():
    out2in_sh = r"""
#!/usr/bin/env bash
out=bbss_out
rm -r -f in
mkdir in
cat $out/tmp > in/tmp
for f in $out/tmp.*.* ; do
  i=`echo "$f" | sed 's/.*tmp\.\([0-9]*\)\..*/\1/'`
  if test ! -f in/tmp.$i ; then
    cnt=`ls $out/tmp.$i.* | wc -l`
    echo $cnt > in/tmp.$i
    cat $out/tmp.$i.* >> in/tmp.$i
  fi
done    
"""
    if pc.id() == 0:
        import tempfile

        with tempfile.NamedTemporaryFile("w") as scriptfile:
            scriptfile.write(out2in_sh)
            scriptfile.flush()
            subprocess.check_call(["/bin/bash", scriptfile.name])

    pc.barrier()


def prun(tstop, mode=None):
    pc.set_maxstep(10 * ms)
    h.finitialize(-65 * mV)

    if mode == "save_test":  # integrate to tstop/2 and save
        pc.psolve(tstop / 2)
        bbss = h.BBSaveState()
        bbss.save_test()  # creates separate file in bbss_out for each piece
    elif mode == "save_test_bin":
        pc.psolve(tstop / 2)
        bbss = h.BBSaveState()
        bbss.save_test_bin()  # like save_test but binary wholecell in binbufout
    elif mode == "save":
        pc.psolve(tstop / 2)
        bbss = h.BBSaveState()
        bbss.save("allcell-bbss.dat")  # single file for everything (nhost==1!)
    elif mode == "restore_test":
        cp_out_to_in()  # prepare for restore.
        bbss = h.BBSaveState()
        bbss.restore_test()
    elif mode == "restore_test_bin":
        subprocess_run("mkdir binbufin")
        subprocess_run("cp binbufout/* binbufin")
        bbss = h.BBSaveState()
        bbss.restore_test_bin()
    elif mode == "restore":
        bbss = h.BBSaveState()
        bbss.restore("allcell-bbss.dat")
        bbss.vector_play_init()  # just for coverage as there are none.

    pc.psolve(tstop)  # integrate the rest of the way


def get_all_spikes(ring):
    local_data = {}
    for i, gid in enumerate(ring.spikegid):
        if gid not in local_data:
            local_data[gid] = []
        local_data[gid].append(ring.spiketime[i])
    all_data = pc.py_allgather([local_data])

    pc.barrier()

    data = {}
    for d in all_data:
        data.update(d[0])

    return data


def compare_dicts(dict1, dict2):
    # assume dict is {gid:[spiketime]}

    # In case iteration order not same in dict1 and dict2, use dict1 key
    # order to access dict
    keylist = dict1.keys()

    # verify same set of keys
    assert set(keylist) == set(dict2.keys())

    # verify same count of spikes for each key
    assert [len(dict1[k]) for k in keylist] == [len(dict2[k]) for k in keylist]

    # Put spike times in array so can compare with a tolerance.
    array_1 = np.array([val for k in keylist for val in dict1[k]])
    array_2 = np.array([val for k in keylist for val in dict2[k]])
    assert np.allclose(array_1, array_2)


def test_3():
    rmfiles()
    hl = hlist()
    npo = hl.count()

    stdspikes = {
        0.0: [
            1.975000000099995,
            21.300000000099324,
            39.70000000010047,
            58.05000000010464,
            76.4000000001088,
            94.80000000011299,
        ],
        1.0: [
            5.9250000001000505,
            24.800000000099125,
            43.32500000010129,
            61.725000000105474,
            80.10000000010965,
            98.47500000011382,
        ],
        2.0: [
            9.900000000099972,
            28.40000000009892,
            46.950000000102115,
            65.3750000001063,
            83.77500000011048,
        ],
        3.0: [
            13.900000000099745,
            32.15000000009875,
            50.600000000102945,
            69.02500000010713,
            87.45000000011132,
        ],
        4.0: [
            17.900000000099517,
            36.02500000009963,
            54.32500000010379,
            72.70000000010796,
            91.12500000011215,
        ],
    }

    ring = Ring(3, 2)
    assert hl.count() == npo

    prun(100 * ms)
    if "usetable_hh" not in dir(h):  # coreneuron has different hh.mod
        stdspikes = get_all_spikes(ring)
    compare_dicts(get_all_spikes(ring), stdspikes)
    stdspikes_after_50 = {}
    for gid in stdspikes:
        stdspikes_after_50[gid] = [spk_t for spk_t in stdspikes[gid] if spk_t >= 50.0]

    prun(100 * ms, mode="save_test")  # at tstop/2 does a BBSaveState.save
    assert hl.count() == npo
    compare_dicts(get_all_spikes(ring), stdspikes)

    prun(100 * ms, mode="restore_test")  # BBSaveState restore to start at t = tstop/2
    assert hl.count() == npo
    compare_dicts(get_all_spikes(ring), stdspikes_after_50)

    # while we are at it check the BBSaveState binary file mode.
    prun(100 * ms, mode="save_test_bin")
    compare_dicts(get_all_spikes(ring), stdspikes)
    prun(100 * ms, mode="restore_test_bin")
    compare_dicts(get_all_spikes(ring), stdspikes_after_50)

    # while we are at it check the BBSaveState save/restore single file mode.
    prun(100 * ms, mode="save")
    compare_dicts(get_all_spikes(ring), stdspikes)
    prun(100 * ms, mode="restore")
    compare_dicts(get_all_spikes(ring), stdspikes_after_50)
    assert hl.count() == npo

    # save_request works but not very useful as save_gid and restore_gid
    # not implemented.
    pc.set_maxstep(10)
    h.finitialize(-65)
    gids = h.Vector()
    sizes = h.Vector()
    bbss = h.BBSaveState()
    ngid = bbss.save_request(gids, sizes)
    assert ngid == len(ring.cells)
    # prints "not implemented"
    bbss.save_gid()
    bbss.restore_gid()
    assert hl.count() == npo
    del bbss

    # binq
    rmfiles()
    h.CVode().queue_mode(1)
    prun(100, mode="save")
    compare_dicts(get_all_spikes(ring), stdspikes)
    prun(100, mode="restore")
    compare_dicts(get_all_spikes(ring), stdspikes_after_50)
    prun(100, mode="save_test_bin")
    compare_dicts(get_all_spikes(ring), stdspikes)
    prun(100, mode="restore_test_bin")
    compare_dicts(get_all_spikes(ring), stdspikes_after_50)
    h.CVode().queue_mode(0)

    pc.gid_clear()


def run2(tstop, mode=None):
    pc.set_maxstep(10)
    h.finitialize()
    if mode == "save_test":
        pc.psolve(tstop / 2)
        h.BBSaveState().save_test()
    elif mode == "restore_test":
        cp_out_to_in()
        h.BBSaveState().restore_test()
    pc.psolve(tstop)


def cmpspikes2(tbegin, tvec, gidvec, tvec_std, gidvec_std):
    i = tvec.indwhere(">=", tbegin)
    j = tvec_std.indwhere(">=", tbegin)
    assert tvec.c(i).eq(tvec_std.c(j))
    assert gidvec.c(i).eq(gidvec_std.c(j))


def test_4():  # for some extra coverage of BBSaveState
    # Save and restore SelfEvent, Random123, NetStims hava a gid
    # NetStim ring
    ncell = 2
    gids = range(pc.id(), ncell, pc.nhost())
    cells = {}
    for gid in gids:
        cells[gid] = h.NetStim()
        cell = cells[gid]
        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(cell, None))
        cell.interval = 1
        cell.number = 3  # >1 has no effect but helps with coverage
        cell.start = 1 if gid == 0 else -1  # gid 0 initiates spike
        cell.noise = 0.1
        cell.noiseFromRandom123(gid, 0, 0)
    netcons = [pc.gid_connect((gid + ncell - 1) % ncell, cells[gid]) for gid in gids]
    for netcon in netcons:
        netcon.weight[0] = 1

    spiketimes = h.Vector()
    spikegids = h.Vector()
    pc.spike_record(-1, spiketimes, spikegids)

    tstop = 20
    run2(tstop)
    spiketimes_std = spiketimes.c()
    spikegids_std = spikegids.c()

    run2(tstop, mode="save_test")
    cmpspikes2(0.0, spiketimes, spikegids, spiketimes_std, spikegids_std)

    run2(tstop, mode="restore_test")
    cmpspikes2(tstop / 2.0, spiketimes, spikegids, spiketimes_std, spikegids_std)

    pc.gid_clear()


def test_5():
    # NetStim (with no gid) stimulates intrinsically firing IntFire2
    # which stimulates a soma with ExpSyn
    gids = range(pc.id(), 1, pc.nhost())
    cells = {}
    for gid in gids:
        cells[gid] = h.IntFire2()
        cell = cells[gid]
        cell.ib = 2
        cell.taus = 2
        cell.taum = 1

        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(cell, None))

    if pc.gid_exists(0):
        ns = h.NetStim()
        nc = h.NetCon(ns, cells[0])
        ns.interval = 1
        ns.start = 0.2
        nc.weight[0] = 0.01

    spiketimes = h.Vector()
    spikegids = h.Vector()
    pc.spike_record(-1, spiketimes, spikegids)

    tstop = 10
    run2(tstop)
    spiketimes_std = spiketimes.c()
    spikegids_std = spikegids.c()

    run2(tstop, mode="save_test")
    cmpspikes2(0.0, spiketimes, spikegids, spiketimes_std, spikegids_std)

    run2(tstop, mode="restore_test")
    cmpspikes2(tstop / 2.0, spiketimes, spikegids, spiketimes_std, spikegids_std)
    for i, t in enumerate(spiketimes):
        print(t, spikegids[i])

    pc.gid_clear()


def test_6():  # Test BBSaveState.ignore
    rmfiles()
    model = Ring(5, 0)
    syn = model.cells[0].syn
    prun(1)
    h.finitialize(-65)
    g_expect = syn.g
    pc.psolve(20)
    g_actual = syn.g
    bbss = h.BBSaveState()
    bbss.ignore(syn)
    bbss.save("allcell-bbss.dat")
    h.finitialize(-65)
    bbss.restore("allcell-bbss.dat")
    print(g_expect, g_actual, syn.g)
    assert syn.g == g_expect
    h.finitialize(-65)
    pc.psolve(20)
    bbss.ignore(syn)
    bbss.ignore()
    bbss.save("allcell-bbss.dat")
    h.finitialize(-65)
    bbss.restore("allcell-bbss.dat")
    print(g_expect, g_actual, syn.g)
    assert abs((syn.g - g_actual) / (syn.g + g_actual)) < 1e-12

    pc.gid_clear()


def test_7():  # save a ring with n cells and extra section, restore n-1 cells
    rmfiles()
    model = Ring(5, 0)
    model.cells[0].foo = h.Section(name="foo", cell=model.cells[0])
    prun(5)
    v_expect = model.cells[0].soma(0.5).v
    bbss = h.BBSaveState()
    bbss.save_test()
    model = Ring(4, 0)
    prun(0.1)
    bbss = h.BBSaveState()
    cp_out_to_in()
    bbss.restore_test()
    print(v_expect, model.cells[0].soma(0.5).v)
    assert abs(v_expect - model.cells[0].soma(0.5).v) < 1e-13

    pc.gid_clear()


def test_8():
    print("focus on BinQ initialization Issue #1444")
    ncell = 3
    n = 5
    # gid -> [n NetCons] -> gid+1
    cells = {gid: h.Follower() for gid in range(pc.id(), ncell, pc.nhost())}
    for gid in cells:
        pc.set_gid2node(gid, pc.id())
        pc.cell(gid, h.NetCon(cells[gid], None))
    netcons = {}
    for gid in cells:
        netcons[gid] = [
            pc.gid_connect(gid - 1, cells[gid]) for _ in range(n) if gid > 0
        ]
    # Stimulate gid 0
    if 0 in cells:
        # stim cell[0] with fast burst of 5 spikes
        ns = h.NetStim()
        ns.start = 0.9999
        ns.number = 5
        ns.interval = 0.1
        nsnc = h.NetCon(ns, cells[0])
        nsnc.delay = 0
        nsnc.weight[0] = 0
    for gid, cell in cells.items():
        # cells generate spike on every input of weight > 1
        # cell.refrac = 0
        for i, nc in enumerate(netcons[gid]):
            nc.weight[0] = 0.001 * i + h.dt
            nc.delay = (1.0 + 0.01 * i) * h.dt
    spiketime = h.Vector()
    spikegid = h.Vector()
    pc.spike_record(-1, spiketime, spikegid)

    def run(tstop):
        pc.set_maxstep(10)
        h.finitialize()
        pc.psolve(tstop)

    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()

    def set_stdspikes():
        spiketime_std.copy(spiketime)
        spikegid_std.copy(spikegid)

    tstop = 5
    run(tstop)
    set_stdspikes()

    def prspikes():
        print("prspikes")
        for i, gid in enumerate(spikegid):
            print("%g %d" % (spiketime[i], gid))

    def compare_spikes():
        x = list(zip(spiketime_std, spikegid_std))
        x = sorted(x, key=lambda e: e[1])
        y = list(zip(spiketime, spikegid))
        y = sorted(y, key=lambda e: e[1])
        if len(x) != len(y):
            print(len(x), len(y))
        # assert len(x) == len(y)
        if x != y:
            q = (x, y) if len(x) <= len(y) else (y, x)
            for i, a in enumerate(q[0]):
                b = q[1][i]
                if a != b:
                    # assert a[1] == b[1]
                    z = abs(a[0] - b[0])
                    if z >= 1e-9:
                        print(x[i], y[i])
                    assert z < 1e-9

    def srun(tsave, tstop):
        cvode = h.CVode()
        qm = cvode.queue_mode()
        print(
            "srun tsave=%g tstop=%g start=%g delay=%g %s"
            % (tsave, tstop, ns.start, nsnc.delay, "binq" if qm else "")
        )
        run(tsave)
        st = spiketime.c()
        sg = spikegid.c()
        bbss = h.BBSaveState()
        bbss.save("allcell_bbss.dat")
        qm = 0
        if qm:
            print("after save")
            h.CVode().print_event_queue()
        h.finitialize(-65)
        spiketime.resize(0)
        spikegid.resize(0)
        bbss.restore("allcell_bbss.dat")
        if qm:
            print("after restore")
            h.CVode().print_event_queue()
        pc.psolve(tstop)
        st.append(spiketime)
        sg.append(spikegid)
        spiketime.copy(st)
        spikegid.copy(sg)
        print("    %d spikes" % len(spiketime))

    srun(1.1, tstop)
    compare_spikes()

    cvode = h.CVode()
    cvode.queue_mode(1)
    run(tstop)
    assert len(spiketime_std) == len(spiketime)
    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()

    srun(1.1, tstop)
    compare_spikes()
    for binq in [0, 1]:
        cvode.queue_mode(binq)
        parms = [
            (1, 0),
            (1e-10, 0),
            (0, 0),
            (0.25 * h.dt, 0),
            (0.5 * h.dt, 0),
            (0.75 * h.dt, 0),
        ]
        for parm in parms:
            if 0 in cells:
                ns.start, nsnc.delay = parm
            for tsave in [0.0, h.dt]:
                run(tstop)
                set_stdspikes()
                srun(tsave, tstop)
                compare_spikes()

    cvode.queue_mode(0)
    pc.gid_clear()


if __name__ == "__main__":
    test_1()
    test_2()
    test_3()
    test_4()
    test_5()
    test_6()
    test_7()
    test_8()



#!/usr/bin/env python
# coding: utf-8

# In[1]:


from neuron import h, rxd

h.load_file("stdrun.hoc")


# In[2]:


axon = h.Section(name="s1")
axon_terminal = h.Section(name="s")
axon.connect(axon_terminal)

axon_terminal_region = rxd.Region([axon_terminal], nrn_region="i")
synaptic_cleft = rxd.Region([axon_terminal], nrn_region="o")
terminal_membrane = rxd.Region(h.allsec(), geometry=rxd.DistributedBoundary(1))


# In[3]:


VA = rxd.Species([axon_terminal_region], name="VA", initial=10, charge=1)
T = rxd.Species([synaptic_cleft], name="T", initial=10, charge=1)


# In[4]:

# if you comment the exocytosis line below, the error will go away
exocytosis = rxd.MultiCompartmentReaction(
    VA[axon_terminal_region],
    T[synaptic_cleft],
    500,
    0,
    membrane=terminal_membrane,
    membrane_flux=True,
)


# In[5]:


VA_conc = h.Vector().record(VA.nodes[0]._ref_concentration)
T_conc = h.Vector().record(T.nodes[0]._ref_concentration)
t_vec = h.Vector().record(h._ref_t)

h.finitialize(-70)
h.continuerun(100)

if __name__ == "__main__":
    from matplotlib import pyplot

    pyplot.plot(t_vec, VA_conc)
    pyplot.plot(t_vec, T_conc)
    pyplot.show()



from neuron import h
from neuron.units import ms, mV
import matplotlib.pyplot as plt
from ring import Ring

cell_to_plot = 0

ring = Ring()

pc = h.ParallelContext()
pc.set_maxstep(10 * ms)

t = h.Vector().record(h._ref_t)
h.finitialize(-65 * mV)
pc.psolve(100 * ms)

if pc.gid_exists(cell_to_plot):
    plt.figure()
    plt.plot(t, pc.gid2cell(cell_to_plot).soma_v)
    plt.show()

pc.barrier()
pc.done()
h.quit()



from neuron import h, rxd
import time

h.load_file("stdrun.hoc")

mM = 1
msec = 1
nM = 1e-6 * mM
sec = 1e3 * msec
minute = 60 * sec
hour = 60 * minute
h.dt = 1 * minute

# def run_sim():
def declare_parameters(**kwargs):
    """enables clean declaration of parameters in top namespace"""
    for key in sorted(kwargs):
        globals()[key] = rxd.Parameter(r, name=key, value=kwargs[key])


def declare_species(**kwargs):
    """enables clean declaration of species in top namespace"""
    for key in sorted(kwargs):
        globals()[key] = rxd.Species(r, name=key, initial=kwargs[key], atolscale=1 * nM)


def recorder(ptr):
    """return a vector that records the pointer"""
    vec = h.Vector()
    vec.record(ptr)
    return vec


rxd.set_solve_type(dimension=3)
cell = h.Section(name="cell")
cell.diam = cell.L = 5
r = rxd.Region(h.allsec())
rxd.nthread(4)

declare_parameters(
    vsP=1.1 * nM / hour,
    vmP=1.0 * nM / hour,
    KmP=0.2 * nM,
    KIP=1.0 * nM,
    ksP=0.9 / hour,
    vdP=2.2 * nM / hour,
    KdP=0.2 * nM,
    vsT=1.0 * nM / hour,
    vmT=0.7 * nM / hour,
    KmT=0.2 * nM,
    KIT=1.0 * nM,
    ksT=0.9 / hour,
    vdT=3.0 * nM / hour,
    KdT=0.2 * nM,
    kdC=0.01 * nM / hour,
    kdN=0.01 * nM / hour,
    k1=0.8 / hour,
    k2=0.2 / hour,
    k3=1.2 / (nM * hour),
    k4=0.6 / hour,
    kd=0.01 * nM / hour,
    V1P=8.0 * nM / hour,
    V1T=8.0 * nM / hour,
    V2P=1.0 * nM / hour,
    V2T=1.0 * nM / hour,
    V3P=8.0 * nM / hour,
    V3T=8.0 * nM / hour,
    V4P=1.0 * nM / hour,
    V4T=1.0 * nM / hour,
    K1P=2.0 * nM,
    K1T=2.0 * nM,
    K2P=2.0 * nM,
    K2T=2.0 * nM,
    K3P=2.0 * nM,
    K3T=2.0 * nM,
    K4P=2.0 * nM,
    K4T=2.0 * nM,
    n=4,
)

declare_species(
    MP=0.0614368 * nM,
    P0=0.0169928 * nM,
    P1=0.0141356 * nM,
    P2=0.0614368 * nM,
    MT=0.0860342 * nM,
    T0=0.0217261 * nM,
    T1=0.0213384 * nM,
    T2=0.0145428 * nM,
    C=0.207614 * nM,
    CN=1.34728 * nM,
)

MTtranscription = rxd.Rate(MT, vsT * KIT**n / (KIT**n + CN**n))
MPtranscription = rxd.Rate(MP, vsP * KIP**n / (KIP**n + CN**n))
MTdegradation = rxd.Rate(MT, -(vmT * MT / (KmT + MT) + kd * MT))
MPdegradation = rxd.Rate(MP, -(vmP * MP / (KmP + MP) + kd * MP))
T0production = rxd.Rate(T0, ksT * MT)
T0degradation = rxd.Rate(T0, -kd * T0)
T1degradation = rxd.Rate(T1, -kd * T1)
T2degradation = rxd.Rate(T2, -kd * T2)
T2degradation_due_to_light = rxd.Rate(T2, -vdT * T2 / (KdT + T2))
T0toT1 = rxd.Reaction(
    T0, T1, V1T * T0 / (K1T + T0), V2T * T1 / (K2T + T1), custom_dynamics=True
)
T1toT2 = rxd.Reaction(
    T1, T2, V3T * T1 / (K3T + T1), V4T * T2 / (K4T + T2), custom_dynamics=True
)
P0production = rxd.Rate(P0, ksP * MP)
P0degradation = rxd.Rate(P0, -kd * P0)
P1degradation = rxd.Rate(P1, -kd * P1)
P2degradation = rxd.Rate(P2, -kd * P2 - vdP * P2 / (KdP + P2))
P0toP1 = rxd.Reaction(
    P0, P1, V1P * P0 / (K1P + P0), V2P * P1 / (K2P + P1), custom_dynamics=True
)
P1toP2 = rxd.Reaction(
    P1, P2, V3P * P1 / (K3P + P1), V4P * P2 / (K4P + P2), custom_dynamics=True
)
P2T2toC = rxd.Reaction(
    P2 + T2, C, 0 + k3, 0 + k4
)  # the 0+ works around a bug prior to NEURON 7.7
CtoCN = rxd.Reaction(
    C, CN, 0 + k1, 0 + k2
)  # the 0+ works around a bug prior to NEURON 7.7
Cdegradation = rxd.Rate(C, -kdC * C)
CNdegradation = rxd.Rate(CN, -kdN * CN)
t = recorder(h._ref_t)
h.finitialize(-65)

mpvec = recorder(MP.nodes[0]._ref_concentration)
cnvec = recorder(CN.nodes[0]._ref_concentration)
p0vec = recorder(P0.nodes[0]._ref_concentration)
p1vec = recorder(P1.nodes[0]._ref_concentration)
p2vec = recorder(P2.nodes[0]._ref_concentration)
cvec = recorder(C.nodes[0]._ref_concentration)

h.finitialize(-65)
h.CVode().active(True)
print("starting simulation")
start = time.time()
h.continuerun(72 * hour)
finish = time.time() - start
print(finish)

pt = (p0vec.as_numpy() + p1vec + p2vec + cvec + cnvec) / nM
mp = mpvec.as_numpy() / nM
cn = cnvec.as_numpy() / nM
t_in_hours = t.as_numpy() / hour

if __name__ == "__main__":
    from matplotlib import pyplot as plt

    plt.figure(figsize=(8, 6))
    plt.plot(t_in_hours, mp, label="MP")
    plt.plot(t_in_hours, cn, label="CN")
    plt.plot(t_in_hours, pt, label="PT")
    plt.legend()
    plt.xlabel("t (hours)")
    plt.ylabel("concentration (nM)")
    plt.show()
